import threading
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .execution.networkrun import NetworkRun


thread_local = threading.local()
thread_local.current_run = None


def set_current_run(current_run: 'NetworkRun'):
    thread_local.current_run = current_run


def get_current_run() -> 'Optional[NetworkRun]':
    current_run = getattr(thread_local, 'current_run', None)
    return current_run


# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains all of the core components of fastr. It has the classes
to create networks and work with them.
"""

from .vfs import VirtualFileSystem


vfs_plugin = VirtualFileSystem()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fastr import exceptions
from fastr.core.ioplugin import IOPlugin

import requests


CHUNK_SIZE = 524288


class HTTPPlugin(IOPlugin):
    """
    .. warning::
        This Plugin is still under development and has not been tested at all.
        example url: https://server.io/path/to/resource
    """
    scheme = ('https', 'http')

    def __init__(self):
        super(HTTPPlugin, self).__init__()

    def fetch_url(self, inurl, outpath):
        """
        Download file from server.

        :param inurl: url to the file.
        :param outpath: path to store file
        """
        response = requests.get(inurl, allow_redirects=True, stream=True, timeout=60)

        # For download (GET) response should be 200
        if response.status_code != 200:
            raise exceptions.FastrDataTypeValueError(
                "Problem downloading the file, HTTP status code {}".format(response.status_code)
            )

        # Write download stream to file
        with open(outpath, 'wb') as output_file:
            for chunk in response.iter_content(CHUNK_SIZE):
                output_file.write(chunk)

        return outpath



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from fastr.data import url
from fastr.datatypes import URLType
from fastr.helpers.checksum import hashsum


class FilePrefix(URLType):
    description = 'Prefix for another file, including the path'
    extension = None  # Explicitly have no extension

    def _validate(value):
        value = value.value

        try:
            if url.isurl(value):
                value = url.get_path_from_url(value)
        except ValueError:
            return False

        return os.path.isdir(os.path.dirname(value))

    def _parse(self):
        return self.value

    def checksum(self):
        return hashsum(self.value)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Job class and some related classes.
"""

import datetime
import os
import urllib.request
import urllib.error
import urllib.parse
import shutil
from typing import Dict, List, Iterable
from collections import OrderedDict
from pathlib import Path

from enum import Enum
import isodate

from .. import exceptions, version, resources
from ..abc.serializable import Serializable, save, load
from ..core import vfs_plugin as vfs
from ..core.cardinality import create_cardinality
from ..core.samples import SampleId, SampleIndex, SampleItem
from ..core.provenance import Provenance
from ..data import url
from ..datatypes import URLType, DataType, Deferred, types
from ..helpers import log, config
from ..helpers.events import emit_event, EventType
from ..helpers.filesynchelper import FileSyncHelper, filesynchelper_enabled
from ..helpers.iohelpers import load_json, save_json

try:
    from fastr.execution.environmentmodules import EnvironmentModules
    ENVIRONMENT_MODULES = EnvironmentModules(config.protected_modules)
    ENVIRONMENT_MODULES_LOADED = True
except exceptions.FastrValueError:
    ENVIRONMENT_MODULES_LOADED = False


class JobState(Enum):
    """
    The possible states a Job can be in. An overview of the states and the
    adviced transitions are depicted in the following figure:

    .. graphviz::

       digraph jobstate {
           nonexistent [shape=box];
           created [shape=box];
           queued [shape=box];
           hold [shape=box];
           running [shape=box];
           execution_done [shape=box];
           execution_failed [shape=box];
           processing_callback [shape=box];
           finished [shape=box];
           failed [shape=box];
           cancelled [shape=box];

           nonexistent -> created;
           created -> queued;
           created -> hold;
           hold -> queued;
           queued -> running;
           running -> execution_done;
           running -> execution_failed;
           execution_done -> processing_callback;
           execution_failed -> processing_callback;
           processing_callback -> finished;
           processing_callback -> failed;
           running -> cancelled;
           queued -> cancelled;
           hold -> cancelled;
       }
    """
    nonexistent = ('nonexistent', 'idle', False)
    created = ('created', 'idle', False)
    queued = ('queued', 'idle', False)
    hold = ('hold', 'idle', False)
    running = ('running', 'in_progress', False)
    execution_done = ('execution_done', 'in_progress', False)
    execution_failed = ('execution_failed', 'in_progress', True)
    execution_skipped = ('execution_skipped', 'in_progress', True)
    processing_callback = ('processing_callback', 'in_progress', False)
    finished = ('finished', 'done', False)
    failed = ('failed', 'done', True)
    cancelled = ('cancelled', 'done', True)

    def __init__(self, _, stage, error):
        self.stage = stage
        self.error = error

    @property
    def idle(self):
        return self.stage == 'idle'

    @property
    def done(self):
        return self.stage == 'done'

    @property
    def in_progress(self):
        return self.stage == 'in_progress'


class JobCleanupLevel(Enum):
    """
    The cleanup level for Jobs that are finished.
    """
    all = 'all'
    non_failed = 'non_failed'
    no_cleanup = 'no_cleanup'


class Job(Serializable):

    """Class describing a job.

       Arguments:
       tool_name - the name of the tool (str)
       tool_version - the version of the tool (Version)
       argument - the arguments used when calling the tool (list)
       tmpdir - temporary directory to use to store output data
       hold_jobs - list of jobs that need to finished before this job can run (list)
    """

    # Constants for file names related to jobs
    COMMAND_DUMP = '__fastr_command__.yaml'
    RESULT_DUMP = '__fastr_result__.yaml'
    STDOUT_DUMP = '__fastr_stdout__.txt'
    STDERR_DUMP = '__fastr_stderr__.txt'
    INFO_DUMP = '__fastr_extra_job_info__.yaml'
    PROV_DUMP = '__fastr_prov__.json'

    def __init__(self,
                 node,
                 sample_id: SampleId,
                 sample_index: SampleIndex,
                 input_arguments: Dict[str, SampleItem],
                 output_arguments: Dict[str, Dict],
                 hold_jobs: List[str]=None,
                 preferred_types: List=None):
        """
        Create a job

        :param fastr.planning.node.Node node: the node the job is based on
        :param sample_id: the id of the sample
        :param sample_index: the index of the sample
        :param input_arguments: the argument list
        :param output_arguments: the argument list
        :param hold_jobs: the jobs on which this jobs depend
        :param preferred_types: The list of preferred types to use
        :return:
        """
        # Save information about the Job environment in which it was created
        self.network_id = node.parent.long_id
        self.network_version = str(node.parent.network_version)
        self.network_scope = node.parent.network_scope
        self.run_id = node.parent.id
        self.node = node
        self.node_id = node.id
        self.node_global_id = node.global_id
        self.tool_id = node.tool.ns_id
        self.tool_version = str(node.tool.version)
        self.sample_id = sample_id
        self.sample_index = sample_index
        self.network_tmpurl = node.parent.tmpurl

        # Arguments for the command
        self.input_arguments = input_arguments
        self.output_arguments = output_arguments

        # Create tmpdir
        # Determine subfolder name in tmp mount
        self.ensure_tmp_dir()

        self._resources = node.resources
        self.translated_values = {}
        self.preferred_types = preferred_types if preferred_types else {}

        if isinstance(hold_jobs, (set, list, tuple)):
            self.hold_jobs = list(hold_jobs)
        elif isinstance(hold_jobs, str):
            self.hold_jobs = [hold_jobs]
        elif hold_jobs is None:
            self.hold_jobs = []
        else:
            raise exceptions.FastrTypeError('Cannot create jobs: hold_jobs has invalid type!')

        self.timestamp = datetime.datetime.utcnow()

        self.info_store = {
            'id': self.id,
        }

        self.errors = []
        self.input_hash = {}
        self.output_hash = {}

        # The status list contains tuples with (timestamp, status)
        # This is updated via the status property
        self.status_list = []
        self.status = JobState.created

        self.provenance = Provenance()

        # Dictionary where the output data will be stored
        self.output_data = {}

        # Save fastr version for safety
        self.fastr_version = version.full_version

    def __getstate__(self):
        """
        Get the state of the job

        :return: job state
        :rtype: dict
        """
        input_arguments = {}
        for key, value in self.input_arguments.items():
            if isinstance(value, SampleItem):
                value = {
                    "index": value.index.__getnewargs__(),
                    "id": value.id.__getnewargs__(),
                    "data": value.data.__getstate__(),
                    "jobs": list(value.jobs),
                    "failed_annotations": list(value.failed_annotations),
                    "status": value.status.name,
                }
            else:
                # FIXME: This causes issues!
                value = str(value)

            input_arguments[key] = value

        # Serialize output data
        output_data = {}
        for key, value in self.output_data.items():
            if isinstance(value, (list, tuple)):
                output_data[key] = [x.serialize() for x in value]
            else:
                output_data[key] = {k: [x.serialize() for x in v] for k, v in value.items()}

        state = {
            # IDs relate to the job
            'network_id': self.network_id,
            'network_version': self.network_version,
            'network_scope': self.network_scope,
            'run_id': self.run_id,
            'node_id': self.node_id,
            'node_global_id': self.node_global_id,
            'tool_id': self.tool_id,
            'tool_version': self.tool_version,
            'sample_id': self.sample_id.__getnewargs__(),
            'sample_index': self.sample_index.__getnewargs__(),

            # Run information for the job
            'fastr_version': self.fastr_version,
            'network_tmpurl': self.network_tmpurl,
            'hold_jobs': self.hold_jobs,
            'timestamp': self.timestamp.isoformat(),
            'preferred_types': {k: [x.id for x in v]
                                for k, v in self.preferred_types.items()},

            # Execution information
            'errors': self.errors,
            'input_hash': self.input_hash,
            'output_hash': self.output_hash,
            'status_list': [{'timestamp': x.isoformat(), 'status': y.name} for x, y in self.status_list],
            'output_data': output_data,

            # Input arguments
            'input_arguments': input_arguments,

            # Output arguments
            'output_arguments': {k: v for k, v in self.output_arguments.items()},
        }

        # Store states by id, not by actual type
        return state

    def __setstate__(self, state):
        """
        Set the state of the job

        :param dict state:
        """
        # Rename argument for backwards compatibility
        self.network_scope = None

        if 'tool_name' in state:
            state['tool_id'] = state.pop('tool_name')

        # Translate sample ID and index
        sample_id = state.get('sample_id')
        if isinstance(sample_id, (list, tuple)):
            state['sample_id'] = SampleId(*sample_id)
        sample_index = state.get('sample_index')
        if isinstance(sample_index, (list, tuple)):
            state['sample_index'] = SampleIndex(*sample_index)

        # Translate timestamp
        timestamp = state.get('timestamp')
        if isinstance(timestamp, str):
            state['timestamp'] = isodate.parse_datetime(timestamp)

        # Translate preferred types to actual types
        preferred_types = {}
        for key, value in state['preferred_types'].items():
            temp = []
            for type_ in value:
                if isinstance(type_, str) and type_ in types:
                    temp.append(types[type_])
                elif isinstance(type_, type) and issubclass(type_, DataType):
                    temp.append(type_)
                else:
                    log.warning('Could not find type {} during unpickling of Job'.format(type_))
            preferred_types[key] = temp

        state['preferred_types'] = preferred_types

        # Translate input arguments
        input_arguments = {}
        for key, input_argument in state['input_arguments'].items():
            if isinstance(input_argument, dict):
                input_argument = SampleItem(**input_argument)
            input_arguments[key] = input_argument
        state['input_arguments'] = input_arguments

        # Translate output data
        output_data = {}
        for key, value in state['output_data'].items():
            if isinstance(value, (list, tuple)):
                output_data[key] = [DataType.deserialize(x) for x in value]
            else:
                output_data[key] = {k: [DataType.deserialize(x) for x in v] for k, v in value.items()}
        state['output_data'] = output_data

        # For backwards compatibility, make sure status list in in correct format
        if 'status_list' not in state and '_status_list' in state:
            state['status_list'] = state.pop('_status_list')

        status_list = []
        for item in state['status_list']:
            if isinstance(item, dict):
                timestamp = item['timestamp']
                if isinstance(timestamp, str):
                    timestamp = isodate.parse_datetime(timestamp)

                status = item['status']
                if isinstance(status, str):
                    status = JobState[status]

                status_list.append([timestamp, status])
            else:
                status_list.append(item)

        state['status_list'] = status_list

        self.__dict__.update(state)

        # Cannot transport reference to Node
        if not hasattr(self, 'node'):
            self.node = None

        if not hasattr(self, 'info_store'):
            self.info_store = {
                'id': self.id,
            }

        if not hasattr(self, 'provenance'):
            self.provenance = Provenance()

    def get_result(self):
        """
        Get the result of the job if it is available. Load the output file if
        found and check if the job matches the current object. If so, load and
        return the result.

        :returns: Job after execution or None if not available
        :rtype: Job | None
        """

        if not self.logfile.exists():
            return None

        log.debug('Found old job result file: {}'.format(self.logfile))

        try:
            result = load(self.logfile)
        except (IOError, EOFError):
            # Errors loading pickle or gzip stream
            return None

        log.debug('Loaded old job result file')

        if not isinstance(result, Job):
            log.debug('Result is not valid Job! (found {})'.format(type(result).__name__))
            return None

        if result.status != JobState.execution_done:
            log.debug('Result status is wrong ({})'.format(result.status))
            return None

        if result.id != self.id:
            log.debug('Result job id is wrong ({})'.format(result.id))
            return None

        if result.tool_id != self.tool_id:
            log.debug('Result tool name is wrong ({})'.format(result.tool_id))
            return None

        if result.tool_version != self.tool_version:
            log.debug('Result tool version is wrong ({})'.format(result.tool_version))
            return None

        if result.sample_id != self.sample_id:
            log.debug('Result sample id is wrong ({})'.format(result.sample_id))
            return None

        log.debug('Checking payloads')

        result_payload = result.create_payload()
        if result_payload != self.create_payload():
            log.debug('Result payload is wrong ({})'.format(result_payload))
            return None

        log.debug('Checking sample index')

        if self.sample_index != result.sample_index:
            log.info('Updating sample index from {} to {}'.format(result.sample_index,
                                                                        self.sample_index))
            result.sample_index = self.sample_index

        # Update job run id to current run
        if self.run_id != result.run_id:
            result.run_id = self.run_id
            result.node_global_id = self.node_global_id

        return result

    def __repr__(self):
        """
        String representation of the Job
        """
        return '<Job\n  id={job.id}\n  tool={job.tool_id} {job.tool_version}\n  tmpdir={job.tmpurl}/>'.format(
            job=self
        )

    @property
    def status(self):
        """
        The status of the job
        """
        # The get last status update and only the new status
        return self.status_list[-1][1]

    @status.setter
    def status(self, status):
        """
        Set the status of a job
        :param status: new status
        """
        if not isinstance(status, JobState):
            raise exceptions.FastrTypeError(
                'Job status should be of class JobState, found [{}] {}'.format(type(status).__name__,
                                                                               status)
            )
        if len(self.status_list) == 0 or self.status != status:
            self.status_list.append((datetime.datetime.utcnow(), status))

            # Emit event that this job has been updated
            emit_event(EventType.job_updated, self)

    @property
    def id(self):
        """
        The id of this job
        """
        return '{}___{}___{}'.format(self.network_id,
                                     self.node_id,
                                     self.sample_id)

    @property
    def fullid(self):
        """
        The full id of the job
        """
        return self.id

    @property
    def tmpurl(self):
        """
        The URL of the tmpdir to use
        """
        return url.join(self.network_tmpurl, self.node_id, str(self.sample_id))

    @property
    def commandurl(self):
        """
        The url of the command pickle
        """
        return url.join(self.tmpurl, self.COMMAND_DUMP)

    @property
    def logurl(self):
        """
        The url of the result pickle
        """
        return url.join(self.tmpurl, self.RESULT_DUMP)

    @property
    def stdouturl(self):
        """
        The url where the stdout text is saved
        """
        return url.join(self.tmpurl, self.STDOUT_DUMP)

    @property
    def stderrurl(self):
        """
        The url where the stderr text is saved
        """
        return url.join(self.tmpurl, self.STDERR_DUMP)

    @property
    def provurl(self):
        """
        The url where the prov document is saved
        """
        return url.join(self.tmpurl, self.PROV_DUMP)

    @property
    def extrainfourl(self):
        """
        The url where the extra job info document is saved
        """
        return url.join(self.tmpurl, self.INFO_DUMP)

    @property
    def resources(self):
        """
        The compute resources required for this job
        """
        return self._resources

    def clean(self):
        log.debug('Cleaning {}'.format(self.id))
        cleanup_level = JobCleanupLevel(config.job_cleanup_level)

        if cleanup_level == JobCleanupLevel.no_cleanup:
            return

        if cleanup_level == JobCleanupLevel.all or self.status not in [JobState.failed, JobState.execution_failed]:
            if self.tmpdir.exists():
                shutil.rmtree(self.tmpdir)

    def get_deferred(self, output_id, cardinality_nr, sample_id=None):
        """
        Get a deferred pointing to a specific output value in the Job

        :param str output_id: the output to select from
        :param int cardinality_nr: the index of the cardinality
        :param str sample_id: the sample id to select (optional)
        :return: The deferred
        """
        parsed_url = urllib.parse.urlparse(self.logurl)

        query = {
            'outputname': output_id,
            'nr': cardinality_nr,
        }

        if sample_id is not None:
            query['sampleid'] = sample_id

        deffered_url = urllib.parse.urlunparse(
            (
                'val',
                parsed_url.netloc,
                parsed_url.path,
                parsed_url.params,
                urllib.parse.urlencode(query),
                ''
            )
        )

        return Deferred(deffered_url)

    @property
    def tmpdir(self) -> Path:
        """
        Path of tempdir for the job
        """
        return Path(vfs.url_to_path(self.tmpurl))

    @property
    def commandfile(self) -> Path:
        """
        The path of the command pickle
        """
        return Path(vfs.url_to_path(self.commandurl))

    @property
    def logfile(self) -> Path:
        """
        The path of the result pickle
        """
        return Path(vfs.url_to_path(self.logurl))

    @property
    def stdoutfile(self) -> Path:
        """
        The path where the stdout text is saved
        """
        return Path(vfs.url_to_path(self.stdouturl))

    @property
    def stderrfile(self) -> Path:
        """
        The path where the stderr text is saved
        """
        return Path(vfs.url_to_path(self.stderrurl))

    @property
    def provfile(self) -> Path:
        """
        The path where the prov document is saved
        """
        return Path(vfs.url_to_path(self.provurl))

    @property
    def extrainfofile(self) -> Path:
        """
        The path where the extra job info document is saved
        """
        return Path(vfs.url_to_path(self.extrainfourl))

    @property
    def tool(self):
        return resources.tools[self.tool_id, self.tool_version]

    def ensure_tmp_dir(self):
        if not self.tmpdir.exists():
            self.tmpdir.mkdir(parents=True, exist_ok=True)

        if not self.tmpdir.exists():
            log.critical('Could not create output directory {}!'.format(self.tmpdir))

    def create_payload(self):
        """
        Create the payload for this object based on all the input/output
        arguments

        :return: the payload
        :rtype: dict
        """
        tool = self.tool
        payload = {'inputs': {}, 'outputs': {}}

        # Fill the payload with the values to use (these should be translated to paths/strings/int etc
        # Translate all inputs to be in correct form
        for id_, value in self.input_arguments.items():
            argument = tool.inputs[id_]
            argument_datatype = types[argument.datatype]
            if isinstance(value, SampleItem):
                if len(value.data.mapping_part()) == 0:
                    value = value.data.sequence_part()
                elif len(value.data.sequence_part()) == 0:
                    value = value.data.mapping_part()
                else:
                    raise ValueError('Fastr does not (yet) accept mixed sequence/mapping input!')

            if not argument.hidden:
                if isinstance(value, tuple):
                    payload['inputs'][id_] = tuple(self.translate_argument(x) for x in value)
                else:
                    # Should be ordered dict
                    payload['inputs'][id_] = OrderedDict(
                        (k, tuple(self.translate_argument(x) for x in v)) for k, v in value.items()
                    )

            else:
                if issubclass(argument_datatype, URLType):
                    payload['inputs'][id_] = tuple(self.translate_argument(x) for x in value)
                else:
                    payload['inputs'][id_] = value

            if len(payload['inputs'][id_]) == 0 and argument.default is not None:
                payload['inputs'][id_] = (self.cast_to_type(argument.default, (argument_datatype,)),)

        # Create output arguments automatically
        for id_, spec in self.output_arguments.items():
            argument = tool.outputs[id_]

            if not argument.automatic:
                cardinality_spec = create_cardinality(spec['cardinality'], None)
                cardinality = cardinality_spec.calculate_job_cardinality(payload)

                if cardinality is None:
                    raise exceptions.FastrValueError('Cardinality cannot be unknown for non-automatic outputs! '
                                                     'Found {} for {}'.format(cardinality, argument))
            else:
                cardinality = 1

            payload['outputs'][id_] = self.fill_output_argument(tool.outputs[id_],
                                                                cardinality,
                                                                spec['datatype'],
                                                                spec['requested'],
                                                                self.tmpurl)

        return payload

    def execute(self):
        """
        Execute this job

        :returns: The result of the execution
        :rtype: InterFaceResult
        """
        # Check if Fastr version is stored or we use a version that predates
        # this feature
        if not hasattr(self, 'fastr_version'):
            message = (
                'Job was created with an old version of Fastr that did not'
                'track the fastr_version yet, current Fastr version is {}'
            ).format(version.full_version)

            # For default branch (production) this is an error, for development
            # just a warning
            if version.git_branch == 'default':
                log.critical(message)
                raise exceptions.FastrVersionMismatchError(message)
            else:
                log.warning(message)

        # Check if the fastr version is identical to the version that created
        # the job initially.
        if version.full_version != self.fastr_version:
            message = (
                'Job was created using Fastr version {}, but execution is'
                ' attempted on Fastr version {}'
            ).format(
                self.fastr_version,
                version.full_version
            )

            # For default branch (production) this is an error, for development
            # just a warning
            if version.git_branch == 'default':
                log.critical(message)
                raise exceptions.FastrVersionMismatchError(message)
            else:
                log.warning(message)

        # Change the working directory to job temp dir
        old_curdir = os.path.abspath(os.curdir)
        job_dir = vfs.url_to_path(self.tmpurl)
        log.info('Set current directory to job output dir {}'.format(job_dir))
        os.chdir(job_dir)

        # Set environment variables
        os.environ['FASTR_SAMPLE_ID'] = str(self.sample_id)
        os.environ['FASTR_SAMPLE_INDEX'] = str(self.sample_index)
        os.environ['FASTR_JOB_ID'] = self.id
        os.environ['FASTR_WORK_DIR'] = job_dir

        tool = self.tool

        # Hash the inputs
        self.hash_inputs()

        # Create the payload
        log.info('Start executing tool')
        start = datetime.datetime.utcnow()
        payload = self.create_payload()
        end = datetime.datetime.utcnow()
        log.info('Finished creating payload in {} seconds'.format((end - start).total_seconds()))

        # Execute the tool
        log.info('Start executing tool')
        start = datetime.datetime.utcnow()
        result = tool.execute(payload)
        end = datetime.datetime.utcnow()
        log.info('Finished executing tool in {} seconds'.format((end - start).total_seconds()))

        # Save the log data
        self.info_store['process'] = result.target_result.as_dict()

        # Check if there were errors in the Interface result
        errors = []
        for error in result.errors:
            if isinstance(error, exceptions.FastrError):
                errors.append(error.excerpt())
            elif isinstance(error, BaseException):
                errors.append((type(error).__name__, error.message, 'unknown', 'uknown'))
            elif isinstance(error, str):
                errors.append(('ErrorInJobExecute', error, 'uknown', 'unknown'))
            elif isinstance(error, (list, tuple)) and len(error) == 4:
                errors.append(error)
            else:
                log.warning('Collected a mal-formatted error: {}'.format(error))
        self.errors.extend(result.errors)

        log.info('Start translating results tool')
        start = datetime.datetime.utcnow()
        self.output_data = self.translate_results(result.result_data)
        end = datetime.datetime.utcnow()
        log.info('Finished translating results in {} seconds'.format((end - start).total_seconds()))

        # Collect the provenance for the node
        self.collect_provenance()

        # Return the working directory to the old state
        log.info('Resetting current directory to {}'.format(old_curdir))
        os.chdir(old_curdir)

        if not self.validate_results(payload):
            raise exceptions.FastrValueError('Output values are not valid!')

        return result

    def collect_provenance(self):
        """
        Collect the provenance for this job.
        """
        self.provenance.collect_provenance(self)

    @classmethod
    def translate_argument(cls, value):
        """
        Translate an argument from a URL to an actual path.

        :param value: value to translate
        :param datatype: the datatype of the value
        :return: the translated value
        """
        return cls.get_value(value=value)

    def get_output_datatype(self, output_id):
        """
        Get the datatype for a specific output

        :param str output_id: the id of the output to get the datatype for
        :return: the requested datatype
        :rtype: tuple
        """
        output = self.tool.outputs[output_id]
        datatype = types[output.datatype]
        preferred_type = datatype
        requested_type = types[self.output_arguments[output_id]['datatype']]

        # If there are preferred types, match with that if possible
        if output_id in self.preferred_types and len(self.preferred_types[output_id]) > 0:
            preferred_type = types.match_types(datatype, preferred=self.preferred_types[output_id])

        datatypes = tuple(x for x in [requested_type, preferred_type, datatype] if x is not None)

        return datatypes

    def translate_results(self, result):
        """
        Translate the results of an interface (using paths etc) to the proper
        form using URI's instead.

        :param dict result: the result data of an interface
        :return: the translated result
        :rtype: dict
        """
        to_remove = []
        for key, value in result.items():
            datatypes = self.get_output_datatype(key)
            try:
                result[key] = self.translate_output_results(value,
                                                            datatypes=datatypes)
            except exceptions.FastrError:
                if self.output_arguments[key]['requested']:
                    raise
                else:
                    log.warning('Encountered problem translating non-requested result for {}, ignoring...'.format(key))
                    to_remove.append(key)

        # Removing keys that were not translated correctly
        for key in to_remove:
            log.info('Dropping non-translated item with key {}'.format(key))
            result.pop(key, None)

        return result

    @staticmethod
    def cast_to_type(value, datatypes: Iterable) -> DataType:
        """
        Try to cast value to one of the given datatypes. Will try all the
        datatypes in order.

        :param tuple datatypes: Possible datatypes to cast to
        :return: casted value
        """
        log.info('Trying to cast {} to one of {}'.format(value, datatypes))
        for datatype in datatypes:
            try:
                item = datatype(str(value))
                if item.valid:
                    log.info('Successfully cast to {!r}'.format(item))
                    return item
                log.info('Could not cast to valid {!r}'.format(item))
            except exceptions.FastrValueError:
                log.info('Could not cast to {!r}'.format(datatype))
        else:
            raise exceptions.FastrValueError('Could not cast "{}" to any of {}'.format(
                value, datatypes
            ))

    @staticmethod
    def translate_output_results(value, datatypes, mountpoint=None):
        """
        Translate the results for on Output

        :param value: the results value for the output
        :param tuple datatypes: tuple of possible datatypes for the output
        :param preferred_type: the preferred datatype of the output
        :return: the update value for the result
        """
        if isinstance(value, dict):
            for subkey, subvalue in value.items():
                new_subvalue = []
                for item in subvalue:
                    if not isinstance(item, DataType):
                        item = Job.cast_to_type(item, datatypes)
                    if isinstance(item, URLType):
                        item.value = vfs.path_to_url(item.value)
                        if filesynchelper_enabled():
                            FileSyncHelper().make_file_promise(item.value)
                    new_subvalue.append(item)
                value[subkey] = tuple(new_subvalue)
        else:
            new_value = []
            for item in value:
                if not isinstance(item, DataType):
                    item = Job.cast_to_type(item, datatypes)
                if isinstance(item, URLType):
                    item.value = vfs.path_to_url(item.value, mountpoint=mountpoint)
                    if filesynchelper_enabled():
                        FileSyncHelper().make_file_promise(item.value)
                new_value.append(item)
            value = new_value

        return value

    @classmethod
    def fill_output_argument(cls, output_spec, cardinality, desired_type, requested, tmpurl):
        """
        This is an abstract class method. The method should take the argument_dict
        generated from calling self.get_argument_dict() and turn it into a list
        of commandline arguments that represent this Input/Output.

        :param int cardinality: the cardinality for this output (can be non for automatic outputs)
        :param DataType desired_type: the desired datatype for this output
        :param bool requested: flag to indicate that the output is requested by Fastr
        :return: the values for this output
        :rtype: list
        """
        values = []

        if not output_spec.automatic and requested:
            # For requested non-automatic outputs, fill them here
            datatype = types[desired_type]

            for cardinality_nr in range(cardinality):
                if datatype.extension is not None:
                    output_url = '{}/{}_{}.{}'.format(tmpurl, output_spec.id, cardinality_nr, datatype.extension)
                else:
                    output_url = '{}/{}_{}'.format(tmpurl, output_spec.id, cardinality_nr)

                # Wrap the output url in the correct DataType
                log.debug('Wrapping {} in a {}'.format(output_url, datatype))
                output_value = datatype(output_url)
                log.debug('Got {}'.format(output_value))

                # Translate to a path and use
                values.append(cls.translate_argument(output_value))
        elif output_spec.automatic:
            datatype = types['Boolean']
            values.append(datatype(requested))

        return tuple(values)

    @classmethod
    def get_value(cls, value):
        """
        Get a value

        :param value: the url of the value
        :param datatype: datatype of the value
        :return: the retrieved value
        """
        if isinstance(value, Deferred):
            value = value.target

        # If the value already has valid datatype, use that and don't guess from scratch
        if isinstance(value, URLType):
            datatype = type(value)
            if url.isurl(str(value)):
                if filesynchelper_enabled():
                    # we are using file promises
                    file_sync_helper = FileSyncHelper()
                    # if we have a file promise for this value
                    # we need to wait until the file is available
                    # or until a timeout has occured
                    if file_sync_helper.has_file_promise(value):
                        file_sync_helper.wait_for_vfs_url(value)
                value = vfs.url_to_path(str(value))
            else:
                if filesynchelper_enabled():
                    # we are using file promises
                    file_sync_helper = FileSyncHelper()
                    # if we have a file promise for this value
                    # we need to wait until the file is available
                    # or until a timeout has occured
                    if file_sync_helper.has_file_promise(value):
                        file_sync_helper.wait_for_file(value)
                if not os.path.exists(str(value)):
                    raise exceptions.FastrValueError(
                        'Found a non-url path ({}) of type {} that does not exist!'.format(value, datatype)
                    )
            return datatype(value)
        elif isinstance(value, DataType):
            return value
        else:
            raise exceptions.FastrTypeError('Arguments should be a DataType, found {}'.format(type(value).__name__))

    def hash_inputs(self):
        """
        Create hashes for all input values and store them in the info store
        """
        for key, value in self.input_arguments.items():
            if value.data.is_sequence:
                self.input_hash[key] = [x.checksum() if x.valid else None for x in value.data.sequence_part()]
            else:
                self.input_hash[key] = {}
                for sample_id, input_val in value.data.mapping_part().items():
                    self.input_hash[key][sample_id] = [x.checksum() if x.valid else None for x in input_val]

    def hash_results(self):
        """
        Create hashes of all output values and store them in the info store
        """
        for output in self.output_arguments.values():
            if not output['requested'] and output['id'] not in self.output_data:
                continue

            id_ = output['id']
            output_value = self.output_data[id_]

            if isinstance(output_value, list):
                self.output_hash[id_] = [x.checksum() if x.valid else None for x in output_value]
            elif isinstance(output_value, dict):
                self.output_hash[id_] = {}
                for sample_id, output_val in output_value.items():
                    self.output_hash[id_][sample_id] = [x.checksum() if x.valid else None for x in output_val]

    def validate_results(self, payload):
        """
        Validate the results of the Job

        :return: flag indicating the results are complete and valid
        """

        valid = True
        for output in self.tool.outputs.values():
            id_ = output.id

            if id_ not in self.output_data:
                if id_ in self.output_arguments and self.output_arguments[id_]['requested']:
                    message = 'Could not find result for output {}'.format(id_)
                    self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
                    log.warning(message)
                    valid = False
                else:
                    log.warning('Could not find non-required output {}'.format(id_))
                continue

            output_value = self.output_data[id_]

            if isinstance(output_value, (list, tuple)):
                if not self._validate_result(output, output_value, payload):
                    message = 'The output "{}" is invalid!'.format(id_)
                    self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
                    log.warning(message)
                    if self.output_arguments[id_]['requested']:
                        valid = False
                    else:
                        log.warning('Non-required output {} is invalid!'.format(id_))
            elif isinstance(output_value, dict):
                for sample_id, output_val in output_value.items():
                    if self._validate_result(output, output_val, payload):
                        continue

                    message = 'The output "{}" for sample "{}" is invalid!'.format(id_, sample_id)
                    self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
                    log.warning(message)
                    if self.output_arguments[id_]['requested']:
                        valid = False
                    else:
                        log.warning('Non-required output {} is invalid!'.format(id_))

            else:
                raise exceptions.FastrTypeError('Output data is not of correct type (expected a list/dict)')

        return valid

    def write(self):
        self.provenance.serialize(self.provfile, 'json')
        save_json(self.extrainfofile, self.info_store)
        save(self, self.logfile)

    def _validate_result(self, output, output_value, payload):
        """
        Validate the result for a specific output/sample
        :param output: the output for which to check
        :param output_value: the value for the output
        :return: flag indicating if the result is value
        """
        valid = True
        cardinality = create_cardinality(output.cardinality, None)

        if not cardinality.validate(payload, len(output_value)):
            message = 'Cardinality mismatch for {} (found {}, expected {!r})'.format(output.id,
                                                                                     len(output_value),
                                                                                     cardinality)
            self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
            log.warning(message)
            valid = False

        for value in output_value:
            if not value.valid:
                message = 'Output value [{}] "{}" not valid for datatype "{!r}"'.format(type(value).__name__,
                                                                                        value,
                                                                                        output.datatype)
                self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
                log.warning(message)
                valid = False

        log.info('Data for output {} is {}'.format(output.id, 'valid' if valid else 'invalid'))

        return valid


class SinkJob(Job):
    """
    Special SinkJob for the Sink
    """
    def __init__(self, node, sample_id, sample_index, input_arguments,
                 output_arguments, hold_jobs=None, substitutions=None,
                 preferred_types=None):

        # This is required!
        self.cardinality = substitutions['cardinality']

        super(SinkJob, self).__init__(node=node,
                                      sample_id=sample_id,
                                      sample_index=sample_index,
                                      input_arguments=input_arguments,
                                      output_arguments=output_arguments,
                                      hold_jobs=hold_jobs,
                                      preferred_types=preferred_types)

        if substitutions:
            self._substitutions = {k: str(v) for k, v in substitutions.items()}
        else:
            self._substitutions = {}

    def __getstate__(self):
        state = super().__getstate__()
        state['substitutions'] = self._substitutions
        return state

    def __setstate__(self, state):
        self._substitutions = state.get('substitutions', {})
        self.cardinality = self._substitutions['cardinality']
        super().__setstate__(state)

    def __repr__(self):
        """
        String representation for the SinkJob
        """
        return '<SinkJob\n  id={job.id}\n  tmpdir={job.tmpurl}/>'.format(job=self)

    @property
    def id(self):
        """
        The id of this job
        """
        return '{}___{}___{}___{}'.format(
            self.network_id,
            self.node_id,
            self.sample_id,
            self.cardinality
        )

    @property
    def tmpurl(self):
        """
        The URL of the tmpdir to use
        """
        return '{}/{}/{}_{}'.format(self.network_tmpurl, self.node_id,
                                    self.sample_id, self.cardinality)

    def get_result(self):
        """
        Get the result of the job if it is available. Load the output file if
        found and check if the job matches the current object. If so, load and
        return the result.

        :returns: Job after execution
        """
        return None

    def validate_results(self, payload):
        """
        Validate the results of the SinkJob

        :return: flag indicating the results are complete and valid
        """
        if self.info_store['process']['stderr'] != '':
            message = 'SinkJob should have an empty stderr, found error messages!\n{}'.format(
                self.info_store['process']['stderr']
            )
            log.warning(message)
            self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
            return False
        else:
            return True

    def create_payload(self):
        """
        Create the payload for this object based on all the input/output
        arguments

        :return: the payload
        :rtype: dict
        """
        payload = super().create_payload()
        log.info('Temp payload: {}'.format(payload))
        log.info('Input data: {}'.format(payload['inputs']['input']))
        log.info('Input arguments: {}'.format(self.input_arguments['input'].data))

        payload['inputs']['output'] = tuple(self.substitute(in_, datatype=type(out)) for in_, out in zip(payload['inputs']['output'],
                                                                                                         payload['inputs']['input']))

        # Add the provenance information into the mix
        extra_inputs = []
        extra_outputs = []
        extra_datatypes = []

        if len(self.input_arguments['input'].data.mapping_part()) > 0:
            raise exceptions.FastrValueError('SinkJob should not have input with mapping_parts!')

        inputs = self.input_arguments['input'].data.sequence_part()
        outputs = payload['inputs']['output']

        print(f'*** PAYLOAD: {payload}')

        for input_, output in zip(inputs, outputs):
            log.info('Checking [{}] {!s} ({!r})'.format(type(input_), input_, input_))
            if isinstance(input_, Deferred):
                parsed_url = urllib.parse.urlparse(input_.raw_value)

                if parsed_url.scheme != 'val':
                    raise exceptions.FastrValueError(
                        'Cannot lookup value {}, wrong url scheme'.format(input_.raw_value)
                    )

                # Find location of the provn file
                datafile = os.path.join(config.mounts[parsed_url.netloc],
                                        os.path.normpath(parsed_url.path[1:]))
                datafile = datafile.replace(self.RESULT_DUMP, self.PROV_DUMP)

                extra_inputs.append(types['JsonFile'](datafile))

                # Update the output_url
                output_url = urllib.parse.urlparse(output)
                output_path = output_url.path
                output_dir, output_file = url.split(output_path)
                output_path = url.join(output_dir, output_file.replace('.', '_')) + '.prov.json'
                output = urllib.parse.urlunparse(output_url._replace(path=output_path))

                extra_outputs.append(output)
                extra_datatypes.append(types['String']('JsonFile'))

        log.info('Extra inputs: {}  Extra outputs: {}'.format(extra_inputs,
                                                              extra_outputs))

        payload['inputs']['input'] = payload['inputs']['input'] + tuple(extra_inputs)
        payload['inputs']['output'] = payload['inputs']['output'] + tuple(extra_outputs)
        payload['inputs']['datatype'] = payload['inputs']['datatype'] + tuple(extra_datatypes)

        return payload

    def substitute(self, value, datatype=None):
        """
        Substitute the special fields that can be used in a SinkJob.

        :param str value: the value to substitute fields in
        :param BaseDataType datatype: the datatype for the value
        :return: string with substitutions performed
        :rtype: str
        """
        if datatype is None:
            datatype = type(value)

        subs = dict(self._substitutions)
        if isinstance(datatype.extension, tuple):
            subs['ext'] = datatype.dot_extension[0]
            subs['extension'] = datatype.extension[0]
        elif datatype.extension is None:
            subs['ext'] = ''
            subs['extension'] = ''
        else:
            subs['ext'] = datatype.dot_extension
            subs['extension'] = datatype.extension

        return str(value).format(**subs)

    def hash_inputs(self):
        """
        Create hashes for all input values and store them in the info store
        """
        input_data = self.input_arguments['input'].data
        if input_data.is_sequence:
            self.input_hash['input'] = [x.checksum() for x in input_data.sequence_part()]
        else:
            self.input_hash['input'] = {}
            for sample_id, input_val in input_data.mapping_part().items():
                self.input_hash['input'][sample_id] = [x.checksum() for x in input_val]


class SourceJob(Job):
    """
    Special SourceJob for the Source
    """

    def __init__(self, datatype, **kwargs):
        super().__init__(**kwargs)
        self.datatype = datatype

    def __getstate__(self):
        state = super().__getstate__()
        state['datatype'] = self.datatype
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self.datatype = state.get('datatype')

    def __repr__(self):
        """
        String representation for the SourceJob
        """
        return '<SourceJob\n  id={job.id}\n  tmpdir={job.tmpurl}/>'.format(job=self)

    def create_payload(self):
        """
        Create the payload for this object based on all the input/output
        arguments

        :return: the payload
        :rtype: dict
        """
        payload = super().create_payload()

        # Inject network scope
        input_part = payload['inputs']
        for key, value in input_part.items():
            new_value = tuple(vfs.expand_network_scope(x, network_scope=self.network_scope) for x in value)
            print(f'New value: {new_value}')
            input_part[key] = new_value

        return payload

    def validate_results(self, payload):
        """
        Validate the results of the Job

        :return: flag indicating the results are complete and valid
        """
        if self.info_store['process']['stderr'] != '':
            message = 'SourceJob should have an empty stderr, found error messages!'
            log.warning(message)
            self.errors.append(exceptions.FastrOutputValidationError(message).excerpt())
            return False
        else:
            return super(SourceJob, self).validate_results(payload)

    def get_output_datatype(self, output_id):
        """
        Get the datatype for a specific output

        :param str output_id: the id of the output to get the datatype for
        :return: the requested datatype
        :rtype: BaseDataType
        """
        return types[self.datatype],

    def hash_inputs(self):
        """
        Create hashes for all input values and store them in the info store
        """
        inputs = [types['String'](x) for x in self.input_arguments['input']]
        self.input_hash['input'] = [x.checksum() for x in inputs]

    def collect_provenance(self):
        """
        Collect the provenance for this job
        """
        super(SourceJob, self).collect_provenance()


class InlineJob(Job):
    """
    Job that does not actually need to run but is used for consistency in
    data processing and logging.
    """
    def __init__(self, *args, **kwargs):
        super(InlineJob, self).__init__(*args, **kwargs)
        self.collect_provenance()

    def collect_provenance(self):
        """
        Collect the provenance for this job
        ::
        """
        self.provenance.init_provenance(self)
        self.provenance.collect_provenance(self, advanced_flow=True)

    def get_result(self):
        if not self.logfile.exists():
            return None

        try:
            result = load_json(self.logfile)
        except (IOError, EOFError):
            # Errors loading pickle or gzip stream
            return None

        if not isinstance(result, Job):
            log.debug('Result is not valid Job! (found {})'.format(type(result).__name__))
            return None

        return result



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The module containing the classes describing the targets.
"""

import imp

from fastr import exceptions, api
from fastr.core.target import Target


class MacroTarget(Target):
    """
    A target for MacroNodes. This target cannot be executed as the MacroNode handles
    execution differently. But this contains the information for the MacroNode to
    find the internal Network.
    """
    def __init__(self, network_file, method=None, function='main'):
        """
        Define a new local binary target. Must be defined either using paths and optionally environment_variables
        and initscripts, or enviroment modules.
        """
        # Figure out required method if not given
        if method is None:
            if network_file.endswith(('.py', '.pyc')):
                method = 'python'
            elif network_file.endswith(('.xml', '.json', 'yml', 'yaml')):
                method = 'loads'

        # Load the network
        if method == 'python':
            network_module = imp.load_source('macro_node.utils', network_file)
            network_function = getattr(network_module, function)
            network = network_function()
        elif method == 'loads':
            network = api.Network.load(network_file)
        else:
            raise exceptions.FastrValueError('Method {} is not know for a MacroTarget'.format(method))

        # Store the network parent
        if not network:
            raise exceptions.FastrValueError('Network not loaded correctly from "{}"'.format(network_file))
        self.network = network.parent

    @classmethod
    def test(cls):
        """
        Test if singularity is availble on the path
        """
        pass

    def run_command(self, command):
        raise exceptions.FastrNotImplementedError(
            'This method is purposefully not implemented, MacroTarget is not mean for direct execution'
        )



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The module containing the classes describing the targets.
"""

from abc import ABCMeta, abstractmethod
from collections import deque, namedtuple
from collections.abc import Sequence
import datetime
import platform
import psutil
import time
import threading
from typing import Dict, List, Union
import subprocess

from .. import exceptions
from ..abc.baseplugin import Plugin
from ..helpers import log, shellescape

SystemUsageInfo = namedtuple('SystemUsageInfo', ['timestamp',
                                                 'cpu_percent',
                                                 'vmem',
                                                 'rmem',
                                                 'read_bytes',
                                                 'write_bytes'])


class TargetResult:
    def __init__(self,
                 return_code: int,
                 stdout: Union[str, bytes],
                 stderr: Union[str, bytes],
                 command: List[Union[str, bytes]],
                 resource_usage: List[SystemUsageInfo],
                 time_elapsed: int
                 ):
        """
        Class to formalize the resulting data of a Target

        :param return_code: the return code of the process
        :param stdout: the stdout generated by the process
        :param stderr: the stderr generated by the process
        :param command: the command executed
        :param resource_usage: the resource use during execution
        :param time_elapsed: time used (in seconds)
        """
        self.return_code = return_code
        self.stdout = stdout if isinstance(stdout, str) else stdout.decode('utf-8')
        self.stderr = stderr if isinstance(stderr, str) else stderr.decode('utf-8')
        self.command = command
        self.resource_usage = resource_usage
        self.time_elapsed = time_elapsed

    def as_dict(self) -> Dict[str, Union[int, str, List]]:
        """
        A dictionary of the data in the object (meant for serialization)
        """
        return {
            "return_code": self.return_code,
            "stdout": self.stdout,
            "stderr": self.stderr,
            "command": self.command,
            "resource_usage": self.resource_usage,
            "time_elapsed": self.time_elapsed,
        }


class ProcessUsageCollection(Sequence):
    # It has to be defined in module for pickling purposes
    usage_type = SystemUsageInfo

    def __init__(self):
        self.seconds_info = deque()
        self.minutes_info = []

    def __len__(self):
        return len(self.seconds_info) + len(self.minutes_info)

    def __getitem__(self, item):
        # First look in minutes, then in seconds
        if item < len(self.minutes_info):
            return self.minutes_info[item]._asdict()
        else:
            return self.seconds_info[item - len(self.minutes_info)]._asdict()

    def append(self, value):
        if not isinstance(value, self.usage_type):
            raise ValueError('Cannot add a non {}.usage_type'.format(type(self).__name__))

        self.seconds_info.append(value)

        if len(self.seconds_info) >= 120:
            self.aggregate(60)

    def aggregate(self, number_of_points):
        oldest_data = [self.seconds_info.popleft() for _ in range(number_of_points)]

        timestamp = oldest_data[-1].timestamp
        cpu_percent = sum(x.cpu_percent for x in oldest_data) / len(oldest_data)
        vmem = max(x.vmem for x in oldest_data)
        rmem = max(x.rmem for x in oldest_data)
        read_bytes = oldest_data[-1].read_bytes
        write_bytes = oldest_data[-1].write_bytes

        self.minutes_info.append(self.usage_type(timestamp=timestamp,
                                                 cpu_percent=cpu_percent,
                                                 vmem=vmem,
                                                 rmem=rmem,
                                                 read_bytes=read_bytes,
                                                 write_bytes=write_bytes))


class Target(Plugin, metaclass=ABCMeta):
    """
    The abstract base class for all targets. Execution with a target should
    follow the following pattern:

    >>> with Target() as target:
    ...     target.run_commmand(['sleep', '10'])

    The Target context operator will set the correct paths/initialization.
    Within the context command can be ran and when leaving the context the
    target reverts the state before.
    """

    # Monitor interval for profiling
    _MONITOR_INTERVAL = 1.0

    def __enter__(self):
        """
        Set the environment in such a way that the target will be on the path.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Cleanup the environment where needed
        """

    @abstractmethod
    def run_command(self, command: List) -> TargetResult:
        """
        Run a command with the target
        """

    @classmethod
    def test(cls):
        """
        Test the plugin, interfaces do not need to be tested on import
        """
        pass


class SubprocessBasedTarget(Target):
    """
    Abstract based class for targets which call the target via a subprocess.
    Supplies a call_subprocess which executes the command and profiles the
    resulting subprocess.
    """
    def call_subprocess(self, command: List) -> TargetResult:
        """
        Call a subprocess with logging/timing/profiling

        :param list command: the command to execute
        :return: execution info
        :rtype: dict
        """
        sysuse = ProcessUsageCollection()
        start_time = time.time()
        log.info('Calling command arguments: {}'.format(command))
        printable_command = []
        for item in command:
            printable_command.append(shellescape.quote_argument(item))
        log.info('Calling command: {}'.format(' '.join(printable_command)))
        try:
            process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        except OSError as exception:
            if exception.errno == 2:
                raise exceptions.FastrExecutableNotFoundError(command[0])
            elif exception.errno == 13:
                # Permission denied
                raise exceptions.FastrNotExecutableError('Cannot execute {}, permission denied!'.format(command[0]))
            else:
                raise exception
        monitor_thread = threading.Thread(target=self.monitor_process, name='SubprocessMonitor', args=(process, sysuse))

        # Make sure this Thread does not block exiting the script
        monitor_thread.daemon = True
        monitor_thread.start()
        stdout, stderr = process.communicate()
        return_code = process.poll()
        end_time = time.time()

        if monitor_thread.is_alive():
            monitor_thread.join(2 * self._MONITOR_INTERVAL)
            if monitor_thread.is_alive():
                log.warning('Ignoring unresponsive monitor thread!')

        return TargetResult(
            return_code=return_code,
            stdout=stdout,
            stderr=stderr,
            command=command,
            resource_usage=list(sysuse),
            time_elapsed=end_time - start_time
        )

    def monitor_process(self, process, resources):
        """
        Monitor a process and profile the cpu, memory and io use. Register the
        resource use every _MONITOR_INTERVAL seconds.

        :param subproces.Popen process: process to monitor
        :param ProcessUsageCollection resources: list to append measurements to
        """
        psproc = psutil.Process(process.pid)
        current_platform = platform.system().lower()

        # Loop initialization
        # Get rid of meaningless 0.0 at start
        psproc.cpu_percent()
        last_timestamp = datetime.datetime.utcnow()

        while process.poll() is None:
            try:
                # The sleep duration is adapted to loop duration so aggregation will not cause
                # extended intervals
                sleep_duration = self._MONITOR_INTERVAL - (datetime.datetime.utcnow() - last_timestamp).total_seconds()
                sleep_duration = 0.0 if sleep_duration < 0.0 else sleep_duration
                time.sleep(sleep_duration)

                # Get process usage information
                memory_info = psproc.memory_info()

                if current_platform == 'darwin':
                    io_read = None
                    io_write = None
                else:
                    try:
                        io_info = psproc.io_counters()
                        io_read = io_info.read_bytes
                        io_write = io_info.write_bytes
                    except psutil.AccessDenied:
                        io_read = None
                        io_write = None

                last_timestamp = datetime.datetime.utcnow()
                usage = resources.usage_type(timestamp=last_timestamp.isoformat(),
                                             cpu_percent=psproc.cpu_percent(),
                                             vmem=memory_info.vms,
                                             rmem=memory_info.rss,
                                             read_bytes=io_read,
                                             write_bytes=io_write)

                resources.append(usage)

            except psutil.Error:
                # If the error occured because during the interval of meassuring the CPU use
                # the process stopped, we do not mind
                if process.poll() is None:
                    raise



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import base64
import subprocess
import threading
import time

import fastr
import fastr.resources
from fastr.plugins.executionplugin import ExecutionPlugin
from fastr.helpers.classproperty import classproperty


class StrongrExecution(ExecutionPlugin):
    _queue = []
    _mappings = {}

    def __init__(self, finished_callback=None, cancelled_callback=None):
        super(StrongrExecution, self).__init__(finished_callback, cancelled_callback)
        fastr.log.debug('Creating strongr job collector')
        self.collector = threading.Thread(name='StrongrJobCollector-0', target=self.check_finished, args=())
        self.collector.daemon = True
        fastr.log.debug('Starting strongr job collector')
        self.collector.start()

    def cleanup(self):
        # Stop thread
        super(StrongrExecution, self).cleanup()

    @classmethod
    def test(cls):
        pass

    @classproperty
    def configuration_fields(cls):
        return {}

    def _job_finished(self, result):
        pass

    def _cancel_job(self, job):
        pass

    def _queue_job(self, job):
        cmd = ['/opt/strongr/addtask', '\'{}\''.format(
            base64.b64encode('/bin/bash -c "{} {} {} {}"'.format(
                '',
                'python',
                 '`python -c \'from fastr.execution import executionscript; print(executionscript.__file__)\'`',
                str(job.commandfile)
            ))), '1', '1']
        print(cmd)
        with open(job.stdoutfile, 'a') as fh_stdout, open(job.stderrfile, 'a') as fh_stderr:
            taskinfo = subprocess.check_output(cmd, stderr=fh_stderr)

        taskid = json.loads(taskinfo)['job_id']
        self._queue.append(taskid)
        self._mappings[taskid] = job

    def check_finished(self):
        while self.running:
            sout = subprocess.check_output('/opt/strongr/queryqueue')
            print(sout)
            queueinfo = json.loads(sout)
            if queueinfo == None:
                time.sleep(5.0)
                continue
            finished = [t for t in self._queue if t not in queueinfo]
            self._queue = [t for t in self._queue if t in queueinfo]

            fastr.log.info('# FINISHED: {}'.format(finished))
            fastr.log.info('# QUEUE: {}'.format(self._queue))

            for taskid in finished:
                fastr.log.info('## TASK ID: {}'.format(taskid))
                self.job_finished(self._mappings[taskid])
            time.sleep(1.0)



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import fastr


def create_macro_network():
    # Create Network
    network = fastr.create_network('add_ints_macro')

    # Create first source
    input_ = network.create_source('Int', id='input')

    # Create calculation nodes
    addint1 = network.create_node('fastr/math/AddInt:1.0', tool_version='1.0', id='addint1', step_id='add')
    addint2 = network.create_node('fastr/math/AddInt:1.0', tool_version='1.0', id='addint2', step_id='add')

    # Link network
    addint1.inputs['left_hand'] = input_.output
    addint1.inputs['right_hand'] = 10,
    addint2.inputs['left_hand'] = addint1.outputs['result']
    addint2.inputs['right_hand'] = 100,

    # Create a sink to save the data
    sink = network.create_sink('Int', id='macro_sink')

    # Link the addint node to the sink
    sink.inputs['input'] = addint2.outputs['result']

    return network


def create_super_macro_node():
    network = fastr.create_network('macro_container')
    # Create Outputs

    input_value = network.create_source('Int', id='input_value')
    # Create Macro Networks

    macro_network_1 = create_macro_network()
    macro_network_2 = create_macro_network()
    add_multiple_ints_node_1 = network.create_macro(macro_network_1, id='node_add_multiple_ints_1')
    add_multiple_ints_node_2 = network.create_macro(macro_network_2, id='node_add_multiple_ints_2')
    # Create Sink

    output_value = network.create_sink('Int', id='output_value')
    # Create Links

    add_multiple_ints_node_1.inputs['input'] = input_value.output
    add_multiple_ints_node_2.inputs['input'] = add_multiple_ints_node_1.outputs['macro_sink']
    output_value.inputs['input'] = add_multiple_ints_node_2.outputs['macro_sink']

    return network


def create_network():
    macro_network = create_super_macro_node()

    # Create Network
    test_network = fastr.create_network('macro_top_level')

    # Create data source
    input_ = test_network.create_source('Int', id='source', step_id='source')

    # Create MacroNode
    add_multiple_ints_node = test_network.create_macro(macro_network, id='node_add_ints')

    # Create sink
    sink = test_network.create_sink('Int', id='sink')

    # Adjust constants(non required inputs) in macro network
    #  add_multiple_ints_node.inputs['const_addint1_value2__add_multiple_ints_1'] = 1234, #input.output
    # Link the network
    add_multiple_ints_node.inputs['input_value'] = input_.output
    sink.inputs['input'] = add_multiple_ints_node.outputs['output_value']

    return test_network


def source_data(network):
    return {'source': [1, 'vfslist://example_data/add_ints/values']}


def sink_data(network):
    return {'sink': 'vfs://tmp/results/{}/result_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id)}


def main():
    network = create_network()

    # Validate and execute network
    if network.is_valid():
        network.draw(network.id, expand_macros=True, draw_dimensions=True)
        network.execute(source_data(network), sink_data(network))
    else:
        print("Network was not valid")


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the NetworkScope plugin for fastr
"""

from fastr.core.ioplugin import IOPlugin


class NetworkScope(IOPlugin):
    """
    A simple source plugin that allows to get data from the Network scope. This uses the
    ``network://`` scheme.

    An uri of ``network://atlases/image_01.nii.gz`` would be translated to
    ``vfs://mount/network/atlases/image_01.nii.gz`` given that the network
    would be created/loaded from ``vfs://mount/network/networkfile.py``.

    .. warning:: This means that the network file must be present in a folder mounted in the
                 ``vfs`` system. Fastr will use a vfs to translate the path between main process
                 and execution workers.

    If the resulting uri should be a different vfs-based url that the default ``vfs://`` then
    a combined scheme can be used. For example ``network+vfslist://atlases/list.txt`` would be
    translated into ``vfslist://mount/network/atlases/list.txt`` and the result would be run
    by the ``vfslist`` plugin.
    """
    scheme = 'network', 'network+'

    def __init__(self):
        # initialize the instance and register the scheme
        super().__init__()



"""
Some reporting functions, e.g. to print a report based on a job result
"""
from . import shellescape


def print_job_result(job_file, print_func=print, verbose=False):
    from fastr.abc.serializable import load
    from fastr.helpers.iohelpers import load_json
    from pprint import pprint

    job = load(job_file)
    try:
        info_store = load_json(job.extrainfofile)
    except FileNotFoundError:
        info_store = {}

    print_func('\n\n===== JOB {} ====='.format(job.id))
    if hasattr(job, 'network_id'):
        print_func('Network: {}'.format(job.network_id))
    if hasattr(job, 'run_id'):
        print_func('Run: {}'.format(job.run_id))

    print_func('Node: {}'.format(job.node_id))
    print_func('Sample index: {}'.format(job.sample_index))
    print_func('Sample id: {}'.format(job.sample_id))
    print_func('Status: {}'.format(job.status))
    print_func('Timestamp: {}'.format(job.timestamp))
    print_func('Job file: {}'.format(job.logfile))

    if hasattr(job, 'errors'):
        errors = job.errors
    elif 'errors' in info_store and isinstance(info_store['errors'], list):
        errors = info_store['errors']
    else:
        errors = []

    print_func('\n----- ERRORS -----')
    for job_error in errors:
        print_func('- {e[0]}: {e[1]} ({e[2]}:{e[3]})'.format(e=job_error))
    print_func('------------------')

    if 'process' in info_store:
        command = info_store['process'].get('command', None)
        if command is not None:
            print_func('\nCommand:')
            print_func('List representation: {}\n'.format(command))
            printable_command = []
            for item in command:
                printable_command.append(shellescape.quote_argument(item))
            print_func('String representation: {}\n'.format(' '.join(printable_command)))
    else:
        print_func('\nNo process information:')
        print_func('Cannot find process information in Job information, processing probably got killed.')
        print_func('If there are no other errors, this is often a result of too high memory use or')
        print_func('exceeding some other type of resource limit.')

    print_func('\nOutput data:')
    pprint(job.output_data)

    print_func('\nStatus history:')
    for timestamp, status in job.status_list:
        print_func('{}: {}'.format(timestamp, status))

    if 'process' in info_store:
        if 'stdout' in info_store['process']:
            print_func('\n----- STDOUT -----')
            print_func(info_store['process']['stdout'])
            print_func('------------------')

        if 'stderr' in info_store['process']:
            print_func('\n----- STDERR -----')
            print_func(info_store['process']['stderr'])
            print_func('------------------')

    if verbose and job.stdoutfile.exists():
        print_func('\n------ FASTR JOB STDOUT -----')
        print_func(job.stdoutfile.read_text())
        print_func('-----------------------------')

    if verbose and job.stderrfile.exists():
        print_func('\n------ FASTR JOB STDERR -----')
        print_func(job.stderrfile.read_text())
        print_func('-----------------------------')



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

import pytest

from sympy.core.symbol import Symbol

from fastr.core.dimension import HasDimensions, Dimension, ForwardsDimensions
from fastr import exceptions


class Dimensional(HasDimensions):
    """
    Test implementation of a HasDimensions class for testing.
    """
    SIZE = (42, 1337, Symbol('X'))
    NAME = ('answer', 'l33t', 'vague')

    def __init__(self):
        self._dimensions = tuple(Dimension(x, y) for x, y in zip(self.NAME, self.SIZE))

    @property
    def dimensions(self):
        return self._dimensions


class DimensionalForwarder(ForwardsDimensions):
    """
    Test the implementation of a ForwardsDimensions class
    """
    def __init__(self, source):
        self._source = source

    @property
    def source(self):
        return self._source

    def combine_dimensions(self, dimensions):
        """
        Strip out last dimension

        :param tuple dimensions: original dimensions
        :return: modified dimensions
        """
        return dimensions[:-1]


"""
Test for the Dimension
"""
@pytest.fixture(scope="module")
def dimension_data():
    data_container = namedtuple('DataContainer', [
        'cleese',
        'cleese2',
        'palin',
        'dummy1',
        'dummy2',
        'jones',
        'unicode',
    ])

    return data_container(
        Dimension('john', 196),
        Dimension('john', 196),
        Dimension('michael', 178),
        Dimension('john', 42),
        Dimension('dummy', 196),
        Dimension('terry', Symbol('length_jones')),
        Dimension('terry\U0001F600', 173)
    )


def test_dimension_repr(dimension_data):
    assert repr(dimension_data.cleese) == '<Dimension john (196)>'
    assert repr(dimension_data.palin) == '<Dimension michael (178)>'


def test_dimension_equal(dimension_data):
    assert dimension_data.cleese == dimension_data.cleese2
    assert dimension_data.cleese != dimension_data.palin
    assert dimension_data.jones != dimension_data.palin

    # Make sure not equal functions
    assert not (dimension_data.cleese != dimension_data.cleese2)

    # Make sure partial correct dimensions do not match
    assert dimension_data.cleese != dimension_data.dummy1
    assert dimension_data.cleese != dimension_data.dummy2


def test_dimension_name(dimension_data):
    assert dimension_data.cleese.name == 'john'
    assert dimension_data.palin.name == 'michael'
    assert dimension_data.jones.name == 'terry'


def test_dimension_size(dimension_data):
    assert dimension_data.cleese.size == 196
    assert dimension_data.palin.size == 178
    assert dimension_data.jones.size == Symbol('length_jones')


def test_dimension_update(dimension_data):
    dimension_data.cleese.update_size(178)
    assert dimension_data.cleese.size == 196

    dimension_data.palin.update_size(196)
    assert dimension_data.palin.size == 196

    dimension_data.jones.update_size(Symbol('length_jones'))
    assert dimension_data.jones.size == Symbol('length_jones')

    dimension_data.jones.update_size(173)
    assert dimension_data.jones.size == 173


def test_dimension_wrong_size_type_str():
    with pytest.raises(exceptions.FastrTypeError):
        Dimension('eric', 'idle')


def test_dimension_wrong_size_type_float():
    with pytest.raises(exceptions.FastrTypeError):
        Dimension('graham', 1.88)


def test_dimension_wrong_value_size():
    with pytest.raises(exceptions.FastrValueError):
        Dimension('negative', -1)


def test_dimension_wrong_update_type(dimension_data):
    with pytest.raises(exceptions.FastrTypeError):
        dimension_data.cleese.update_size(1.96)


def test_dimension_wrong_update_value(dimension_data):
    with pytest.raises(exceptions.FastrValueError):
        dimension_data.jones.update_size(-173)


@pytest.fixture(scope="module")
def dimensional():
        return Dimensional()


def test_dimensiona_size(dimensional):
    assert dimensional.size == dimensional.SIZE


def test_dimensional_dimnames(dimensional):
    assert dimensional.dimnames == dimensional.NAME


def test_dimensional_ndims(dimensional):
    assert dimensional.ndims == 3


@pytest.fixture(scope="module")
def dimensional_forwarder(dimensional):
    return DimensionalForwarder(dimensional)

def test_dimensional_forwarder_size(dimensional, dimensional_forwarder):
    assert dimensional_forwarder.size == dimensional.SIZE[:-1]


def test_dimensional_forwarder_dimnames(dimensional, dimensional_forwarder):
    assert dimensional_forwarder.dimnames == dimensional.NAME[:-1]


def test_dimensional_forwarder_ndims(dimensional, dimensional_forwarder):
    assert dimensional_forwarder.ndims == 2



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module holds the ExecutionPluginManager as well as the base-class
for all ExecutionPlugins.
"""

from .pluginmanager import PluginSubManager
from ...core.interface import Interface


class InterfacePluginManager(PluginSubManager):
    """
    Container holding all the CollectorPlugins
    """

    def __init__(self, parent):
        """
        Create the Coll
        :param path:
        :param recursive:
        :return:
        """
        super(InterfacePluginManager, self).__init__(parent=parent,
                                                     plugin_class=Interface)

    @property
    def _instantiate(self):
        """
        Indicate that the plugins should not instantiated before being stored
        """
        return False



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import urllib.parse as up

from .pluginmanager import PluginSubManager
from ...core.ioplugin import IOPlugin
from ...core.tool import Tool
from ...core.version import Version


class IOPluginManager(PluginSubManager):
    """
    A mapping containing the IOPlugins known to this system
    """

    def __init__(self, parent):
        """
        Create the IOPluginManager and populate it.

        :return: newly created IOPluginManager
        """
        super().__init__(parent=parent,
                         plugin_class=IOPlugin)
        self._key_map = {}

    def cleanup(self):
        """
        Cleanup all plugins, this closes files, connections and other things
        that could be left dangling otherwise.
        """
        for ioplugin in list(self.values()):
            ioplugin.cleanup()

    def __keytransform__(self, key):
        try:
            return self._key_map[key]
        except KeyError:
            self._key_map.clear()
            for id_, value in self.data.items():
                if isinstance(value.scheme, tuple):
                    for scheme in value.scheme:
                        self._key_map[scheme] = id_
                else:
                    self._key_map[value.scheme] = id_

            return self._key_map[key]

    def __iter__(self):
        for value in self.data.values():
            if isinstance(value.scheme, tuple):
                for scheme in value.scheme:
                    yield scheme
            else:
                yield value.scheme

    def _print_key(self, key):
        """
        Get a printable string for the IOPlugin key

        :param key: key to get the printable version for
        :return: printable version of the key
        :rtype: str
        """
        return self[key].status.value, '{}://'.format(key)

    def expand_url(self, url):
        """
        Expand the url by filling the wildcards. This function checks the url scheme
        and uses the expand function of the correct IOPlugin.

        :param str url: url to expand
        :return: list of urls
        :rtype: list of str
        """
        parsed_url = up.urlparse(url)
        return self[parsed_url.scheme].expand_url(url)

    def pull_source_data(self, url, outdir, sample_id, datatype=None):
        """
        Retrieve data from an external source. This function checks the url scheme and
        selects the correct IOPlugin to retrieve the data.

        :param url: url to pull
        :param str outdir: the directory to write the data to
        :param DataType datatype: the datatype of the data, used for determining
                                  the total contents of the transfer
        :return: None
        """
        parsed_url = up.urlparse(url)
        return self[parsed_url.scheme].pull_source_data(url, outdir, sample_id, datatype)

    def push_sink_data(self, inpath, outurl, datatype=None):
        """
        Send data to an external source. This function checks the url scheme and
        selects the correct IOPlugin to retrieve the data.

        :param str inpath: the path of the data to be pushed
        :param str outurl: the url to write the data to
        :param DataType datatype: the datatype of the data, used for determining
                                  the total contents of the transfer
        """
        parsed_url = up.urlparse(outurl)
        return self[parsed_url.scheme].push_sink_data(inpath=inpath,
                                                      outurl=outurl,
                                                      datatype=datatype)

    def put_url(self, inpath, outurl):
        """
        Put the files to the external data store.

        :param inpath: path to the local data
        :param outurl: url to where to store the data in the external data store.
        """
        parsed_url = up.urlparse(outurl)
        return self[parsed_url.scheme].put_url(inpath, outurl)

    def url_to_path(self, url):
        """
        Retrieve the path for a given url

        :param str url: the url to parse
        :return: the path corresponding to the input url
        :rtype: str
        """
        parsed_url = up.urlparse(url)
        return self[parsed_url.scheme].url_to_path(url)

    @staticmethod
    def register_url_scheme(scheme):
        """
        Register a custom scheme to behave http like. This is needed to parse
        all things properly with urlparse.

        :param scheme: the scheme to register
        """
        for method in [s for s in dir(up) if s.startswith('uses_')]:
            if scheme not in getattr(up, method):
                getattr(up, method).append(scheme)

    @staticmethod
    def create_ioplugin_tool(tools, interfaces):
        """
        Create the tools which handles sinks and sources. The command of this tool is the main of core.ioplugin.
        """
        if ('fastr/Source:1.0', '1.0') not in tools:
            source_tool = Tool()
            source_tool._id = 'Source'
            source_tool.name = 'Source'
            source_tool.description = 'Tool to execute Fastr source nodes'
            source_tool.authors = []
            source_tool.version = Version('1.0')
            source_tool.namespace = 'fastr'
            source_tool.node_class = 'SourceNode'
            source_tool.filename = __file__
            source_tool.tests = []
            source_tool._target = None
            source_tool.command = {'authors': ['fastr devs'],
                                   'description': 'Source Tool: tool to handle SourceNode operation.',
                                   'targets': [{
                                       'os': '*',
                                       'arch': '*',
                                       'bin': 'source.py',
                                       'interpreter': sys.executable,
                                       'paths': '../../utils/cmd/',
                                       'module': None
                                   }],
                                   'url': 'http://www.bigr.nl/fastr',
                                   'version': Version('1.0')}

            input_params = [{'id': 'input',
                             'name': 'input',
                             'description': 'The data to be retrieved by the SourceNode',
                             'datatype': 'String',
                             'prefix': '--input',
                             'repeat_prefix': False,
                             'cardinality': '1-*',
                             'required': True,
                             'hidden': True},
                            {'id': 'datatype',
                             'name': 'datatype',
                             'datatype': 'String',
                             'prefix': '--datatype',
                             'cardinality': '1',
                             'required': True,
                             'hidden': True},
                            {'id': 'targetdir',
                             'name': 'targetdir',
                             'description': 'The location to store the result',
                             'datatype': 'Directory',
                             'prefix': '--output',
                             'cardinality': '1',
                             'required': True,
                             'hidden': True},
                            {'id': 'sample_id',
                             'name': 'sample_id',
                             'description': 'The sample id for the SourceJob that is run',
                             'datatype': 'String',
                             'prefix': '--sample_id',
                             'cardinality': '1',
                             'required': True,
                             'hidden': True}]

            output_params = [{'id': 'output',
                              'name': 'The output urls in vfs scheme',
                              'description': '',
                              'datatype': 'AnyType',
                              'cardinality': 'unknown',
                              'automatic': True,
                              'location': '^__IOPLUGIN_OUT__=(.*)$',
                              'method': 'json'}]

            interface = {'inputs': input_params, 'outputs': output_params}
            source_tool.interface = interfaces['FastrInterface'](id_='source-interface', document=interface)

            # Register source tool with the ToolManager
            tools['fastr/Source:1.0', '1.0'] = source_tool

        if ('fastr/Sink:1.0', '1.0') not in tools:
            sink_tool = Tool()
            sink_tool._id = 'Sink'
            sink_tool.name = 'Sink'
            sink_tool.description = 'Tool to execute Fastr sink nodes'
            sink_tool.authors = []
            sink_tool.version = Version('1.0')
            sink_tool.namespace = 'fastr'
            sink_tool.node_class = 'SinkNode'
            sink_tool.filename = __file__
            sink_tool.tests = []
            sink_tool._target = None
            sink_tool.command = {'authors': ['fastr devs'],
                                 'description': 'Sink Tool: tool to handle Sink Node operation.',
                                 'targets': [{
                                     'os': '*',
                                     'arch': '*',
                                     'bin': 'sink.py',
                                     'interpreter': sys.executable,
                                     'paths': '../../utils/cmd/',
                                     'module': None
                                 }],
                                 'url': 'http://www.bigr.nl/fastr',
                                 'version': Version('1.0')}

            input_params = [{'id': 'input',
                             'name': 'The url to process (can also be a list)',
                             'description': 'The data to be store by the SinkNode',
                             'datatype': 'AnyType',
                             'prefix': '--input',
                             'cardinality': '1-*',
                             'required': True, },
                            {'id': 'output',
                             'name': 'output',
                             'datatype': 'String',
                             'prefix': '--output',
                             'cardinality': '1',
                             'required': True,
                             'hidden': True},
                            {'id': 'datatype',
                             'name': 'datatype',
                             'datatype': 'String',
                             'prefix': '--datatype',
                             'cardinality': '1',
                             'required': True,
                             'hidden': True},
                            ]

            output_params = []

            interface = {'inputs': input_params, 'outputs': output_params}
            sink_tool.interface = interfaces['FastrInterface'](id_='sink-interface', document=interface)

            # Register sink tool with the ToolManager
            tools['fastr/Sink:1.0', '1.0'] = sink_tool





#!/usr/bin/python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The executionscript is the script that wraps around a tool executable. It
takes a job, builds the command, executes the command (while profiling it)
and collects the results.
"""

import datetime
import os
import sys
import traceback

print('=== Loading Fastr execution script  ===')
print('Using python {} loaded from {}'.format(sys.version, sys.executable))

FASTR_LOG_TYPE = 'console'
import fastr
from fastr import exceptions
from fastr.execution.job import Job, JobState
from fastr.abc.serializable import load
from fastr.helpers.filesynchelper import FileSyncHelper, filesynchelper_enabled
from fastr.helpers.sysinfo import get_hostinfo, get_sysinfo
from fastr.resources import ioplugins

_MONITOR_INTERVAL = 1.0

# We will run untrusted subprocesses and want to log everything also in case
# of an error. Therefor there are a few bare-except blocks which are intended
# to ensure as much as possible logging in case of errors.
# pyline: disable=bare-except


def execute_job(job):
    """
    Execute a Job and save the result to disk

    :param job: the job to execute
    """
    fastr.log.info('Using Python {}'.format(sys.version))
    fastr.log.info('FASTR loaded from {}'.format(fastr.__file__))
    fastr.log.info('version: {}'.format(fastr.version.full_version))
    fastr.log.info('Start time: {}'.format(datetime.datetime.now()))

    logfile_path = job.logfile
    fastr.log.info('Job log path: {}'.format(logfile_path))

    fastr.log.info('Running job {}\n  command: {} v{}\n  arguments: {}\n  outputs: {}\n'.format(
        job.id, job.tool_id, job.tool_version, job.input_arguments, job.output_arguments)
    )
    try:
        # Checking for old results of a job run before continuing
        old_job_result = job.get_result()

        if old_job_result is not None:
            fastr.log.info('Found a valid job results, re-using that as output')
            job = old_job_result

        else:
            fastr.log.info('No old result, executing job')
            job.status = JobState.running

            # Initialize provenance
            job.provenance.init_provenance(job)

            job.info_store['hostinfo'] = get_hostinfo()
            job.info_store['sysinfo_start'] = get_sysinfo()

            fastr.log.info('DRMAA info: {}'.format(job.info_store['hostinfo']['drmaa']))

            fastr.log.info('Writing intermediate job info to: {}'.format(logfile_path))
            job.write()

            job.execute()

            # Document system information after
            fastr.log.info('Job subprocess finished')
            job.info_store['sysinfo_end'] = get_sysinfo()

            fastr.log.info('Start hashing results')
            start = datetime.datetime.now()
            job.hash_results()
            end = datetime.datetime.now()
            fastr.log.info('Finished hashing results in {} seconds'.format((end - start).total_seconds()))
            fastr.log.info('try end time: {}'.format(datetime.datetime.now()))
    except BaseException as exception:
        fastr.log.warning('Caught exception in execution: [{}] {}'.format(type(exception).__name__,
                                                                          exception))
        job.status = JobState.execution_failed
        traceback.print_exc()

        # Log errors to the info store
        if 'process' not in job.info_store:
            job.errors.append(exceptions.FastrSubprocessNotFinished('There is no information that the subprocess finished properly: appears the job crashed before the subprocess registered as finished.').excerpt())
        elif job.info_store['process']['stderr'] != '':
            job.errors.append(exceptions.FastrErrorInSubprocess(job.info_store['process']['stderr']).excerpt())

        # Add exception as error in job
        exc_type, exception, trace = sys.exc_info()
        exc_info = traceback.format_exc()
        if isinstance(exception, exceptions.FastrError):
            job.errors.append(exception.excerpt())
        else:
            trace = traceback.extract_tb(trace, 1)[0]
            job.errors.append((exc_type.__name__, exception.message, trace[0], trace[1]))
        fastr.log.critical('Execution script encountered errors: {}'.format(exc_info))
    else:
        fastr.log.info('Execution finished normally.')
        job.status = JobState.execution_done
    finally:
        fastr.log.info('Writing job result to: {}'.format(logfile_path))
        job.write()

        # Cleanup all plugins
        ioplugins.cleanup()

        if filesynchelper_enabled():
            FileSyncHelper().job_finished(job.logurl)

        fastr.log.info('End time: {}'.format(datetime.datetime.now()))

        if job.status != JobState.execution_done:
            sys.exit(1)  # Signal that the job failed


def main(joblist=None):
    """
    This is the main code. Wrapped inside a function to avoid the variables
    being seen as globals and to shut up pylint. Also if the joblist argument
    is given it can run any given job, otherwise it takes the first command
    line argument.
    """
    fastr.log.info('----- Execution script -----\n')
    if joblist is None:
        joblist = sys.argv[1]
        if filesynchelper_enabled():
            # we are using file promises
            # running this execution script with a path to
            # a pickle is a promise in itself so we should
            # wait until jobfile exists or timeout occurs
            file_sync_helper = FileSyncHelper()
            file_sync_helper.wait_for_pickle(joblist)

    if os.path.exists(joblist) and os.path.isfile(joblist):
        fastr.log.info('Loading pickled command from file')
        start = datetime.datetime.now()
        joblist = load(joblist)
        end = datetime.datetime.now()
        fastr.log.info('Finished loading pickle in {} seconds'.format((end - start).total_seconds()))

    # Both the error we raise and not able to iterate joblist will result in a
    # TypeError, so we catch both and give our own TypeError message
    try:
        if not isinstance(joblist, Job) and not all(isinstance(x, Job) for x in joblist):
            raise TypeError('Wrong type, found {}!'.format(type(joblist)))
    except TypeError:
        message = 'Argument should be a Job or a iterable of Jobs! (Found: {})'.format(joblist)
        fastr.log.critical(message)
        raise exceptions.FastrTypeError(message)

    fastr.log.info('Received command: {command}\n'.format(command=joblist))

    if isinstance(joblist, list) or isinstance(joblist, tuple):
        for job in joblist:
            execute_job(job)
    elif isinstance(joblist, Job):
        execute_job(joblist)
    fastr.log.info('---------------------------\n')


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The module containing the classes describing the targets.
"""

import os
import time
import threading
from typing import List

import isodate
import requests

import fastr
from fastr import exceptions
from fastr.core.target import Target, ProcessUsageCollection, TargetResult

# Check if docker is available
try:
    import docker
    DOCKER_AVAILABLE = True
except ImportError:
    docker = None
    DOCKER_AVAILABLE = False


class DockerTarget(Target):
    """
    A tool target that is located in a Docker images. Can be run using
    docker-py. A docker target only need two variables: the binary to call within the
    docker container, and the docker container to use.


    .. code-block:: json

        {
          "arch": "*",
          "os": "*",
          "binary": "bin/test.py",
          "docker_image": "fastr/test"
        }

    .. code-block:: xml

        <target os="*" arch="*" binary="bin/test.py" docker_image="fastr/test">

    """
    def __init__(self, binary, docker_image):
        """
        Define a new docker target.

        :param str docker_image: Docker image to use
        """
        if not DOCKER_AVAILABLE:
            raise exceptions.FastrOptionalModuleNotAvailableError('Target cannot be used, module "docker" unavailable')

        self.binary = binary
        self._docker_image = docker_image

        #: Docker api to use for docker target
        self.docker_api = 'unix://var/run/docker.sock'

        self._docker_client = docker.DockerClient(base_url=self.docker_api,
                                                  version='auto')
        self._container = None
        self._running_container = False

    def __enter__(self):
        super(DockerTarget, self).__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        self._container = None

    @property
    def container(self):
        return self._container

    def run_command(self, command: List) -> TargetResult:
        # Bind all fastr mounts
        mounts = list(fastr.config.mounts.values())
        fastr.log.info('DOCKER MOUNTS: {}'.format(mounts))
        binds = {x: {'bind': x, 'mode': 'ro'} for x in mounts if os.path.exists(x)}
        binds[fastr.config.mounts['tmp']]['mode'] = 'rw'
        fastr.log.info('Docker binds: {}'.format(binds))

        fastr.log.info('Creating Docker container')
        container = self._docker_client.containers.run(
            image=self._docker_image,
            entrypoint=command[0],
            command=command[1:],
            detach=True,
            volumes=binds,
            network_disabled=False,
            user='{}:{}'.format(os.getuid(), os.getgid()),
            working_dir=os.path.abspath(os.curdir),
        )

        # Start monitoring
        self._running_container = True
        sysuse = ProcessUsageCollection()
        monitor_thread = threading.Thread(target=self.monitor_docker,
                                          name='DockerMonitor',
                                          args=(container, sysuse))
        monitor_thread.daemon = True  # Make sure this Thread does not block exiting the script
        monitor_thread.start()
        start_time = time.time()

        # Run docker container
        fastr.log.info('Waiting for Docker container to finish')
        return_code = container.wait()
        fastr.log.info('Docker container is done')
        end_time = time.time()
        stdout = container.logs(stdout=True, stderr=False, stream=False, timestamps=False)
        stderr = container.logs(stdout=False, stderr=True, stream=False, timestamps=False)
        self._running_container = False

        if monitor_thread.is_alive():
            monitor_thread.join(2 * self._MONITOR_INTERVAL)
            if monitor_thread.is_alive():
                fastr.log.warning('Ignoring unresponsive monitor thread!')

        # Clean the container
        container.remove()

        return TargetResult(
            return_code=return_code,
            stdout=stdout,
            stderr=stderr,
            command=command,
            resource_usage=list(sysuse),
            time_elapsed=end_time - start_time,
        )

    def monitor_docker(self, container, resources):
        """
        Monitor a docker container and profile the cpu, memory and io use.
        Register the resource use every _MONITOR_INTERVAL seconds.

        :param ContainerCollection container: process to monitor
        :param ProcessUsageCollection resources: list to append measurements to
        """
        try:
            while self._running_container:
                stat = container.stats(stream=False,
                                       decode=False)

                # Get cpu, memory and io statistics
                timestamp = isodate.isodatetime.parse_datetime(stat['read'])
                usage = resources.usage_type(timestamp=timestamp.isoformat(),
                                             cpu_percent=stat['cpu_stats']['cpu_usage']['total_usage'],
                                             vmem=-1.0,
                                             rmem=stat['memory_stats'].get('rss'),
                                             read_bytes=0.0,
                                             write_bytes=0.0)

                resources.append(usage)
                if self._running_container:
                    time.sleep(self._MONITOR_INTERVAL)
        except requests.exceptions.ReadTimeout:
            fastr.log.info('Docker Monitor timed out')



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools
import weakref
from abc import ABCMeta, abstractmethod
from collections import OrderedDict

import fastr
from fastr.core.dimension import HasDimensions
from fastr.core.samples import SampleId, SampleIndex, SampleState
from fastr import exceptions


class BaseInputGroupCombiner(HasDimensions, metaclass=ABCMeta):
    """
    An object that takes the different input groups and combines them in the
    correct way.
    """

    def __init__(self, parent):
        self.parent = weakref.ref(parent)
        self.update()

    @property
    def fullid(self):
        """
        The full id of the InputGroupCombiner
        """
        return '{}/input_group_combiner'.format(self.parent.fullid)

    @property
    def input_groups(self):
        return self.parent().input_groups

    @abstractmethod
    def merge(self, list_of_items):
        """
        Given a list of items for each input group, it returns the combined
        list of items.

        :param list list_of_items: items to combine
        :return: combined list
        """

    @abstractmethod
    def unmerge(self, item):
        """
        Given a item it will recreate the seperate items, basically this is the
        inverse operation of merge. However, this create an OrderedDict so that
        specific input groups can be easily retrieved. To get a round trip, the
        values of the OrderedDict should be taken::

            >>> odict_of_items = combiner.unmerge(item)
            >>> item = combiner.merge(odict_of_items.values())

        :param list item: the item to unmerge
        :return: items
        :rtype: OrderedDict
        """

    @abstractmethod
    def iter_input_groups(self):
        """
        Iterate over all the merged samples
        :return:
        """

    @property
    def dimensions(self):
        dimensions = tuple(self.merge([ig.dimensions for ig in self.input_groups.values()]))
        return dimensions

    def merge_sample_id(self, list_of_sample_ids):
        return SampleId(self.merge(list_of_sample_ids))

    def merge_sample_index(self, list_of_sample_indexes):
        return SampleIndex(self.merge(list_of_sample_indexes))

    def merge_sample_data(self, list_of_sample_data):
        return {k: v if v is not None and len(v) > 0 else None for data in list_of_sample_data for k, v in data.items()}

    def merge_sample_jobs(self, list_of_sample_jobs):
        return set.union(*list_of_sample_jobs)

    def merge_failed_annotations(self, list_of_failed_annotations):
        return set.union(*list_of_failed_annotations)

    def merge_sample_status(self, states):
        return SampleState.combine(states)

    def merge_payloads(self, sample_payloads):
        # Determine the sample index
        sample_index = self.merge_sample_index([x.index for x in sample_payloads])

        # Create sampleid
        sample_id = self.merge_sample_id([x.id for x in sample_payloads])

        # Extract jobdata and combine in single dict
        job_data = self.merge_sample_data(x.data for x in sample_payloads)

        # Create superset of all job dependencies
        job_depends = self.merge_sample_jobs(x.jobs for x in sample_payloads)

        # Create superset of all failed annotations
        failed_annotations = self.merge_sample_jobs(x.failed_annotations for x in sample_payloads)

        # Merge sample states into a single status
        sample_status = self.merge_sample_status(x.status for x in sample_payloads)

        return sample_index, sample_id, job_data, job_depends, failed_annotations, sample_status

    def __iter__(self):
        return self.iter_input_groups()

    def update(self):
        pass


class DefaultInputGroupCombiner(BaseInputGroupCombiner):
    """
    The default input group combiner combines the input group in a cross
    product version, taking each combinations of samples between the input
    groups. So if there are two input groups with one with size N and the
    other with size M x P the result would be N x M x P samples, with all
    possible combinations of the samples in each input group.
    """
    def merge(self, list_of_items):
        """
        Given a list of items for each input group, it returns the combined
        list of items.

        :param list list_of_items: items to combine
        :return: combined list
        """
        return [x for item in list_of_items for x in item]

    def unmerge(self, item):
        """
        Given a item it will recreate the seperate items, basically this is the
        inverse operation of merge. However, this create an OrderedDict so that
        specific input groups can be easily retrieved. To get a round trip, the
        values of the OrderedDict should be taken::

            >>> odict_of_items = combiner.unmerge(item)
            >>> item = combiner.merge(odict_of_items.values())

        :param list item: the item to unmerge
        :return: items
        :rtype: OrderedDict
        """
        result = OrderedDict()
        for target in list(self.input_groups.values()):
            mask = [True if ig.id == target.id else False for ig in self.input_groups.values() for _ in ig.size]
            result[target.id] = tuple(k for k, m in zip(item, mask) if m)

        return result

    def iter_input_groups(self):
        for sample_payloads in itertools.product(*[ig.iterinputvalues for ig in self.input_groups.values()]):
            fastr.log.debug('sample_payload: {}'.format(sample_payloads))
            fastr.log.debug('sample payload data: {}'.format([x.data for x in sample_payloads]))
            yield self.merge_payloads(sample_payloads)


class MergingInputGroupCombiner(BaseInputGroupCombiner):
    """
    The merging input group combiner takes a similar approach as the default
    combiner but merges dimensions that are the same. If input group A has
    N(3) x M(2) samples and B has M(2) x P(4) it wil not result in N(3) x M(2)
    x M(2) x P(4), but merge the dimensions M leading to N(3) x M(2) x P(4) in
    resulting size.
    """
    def __init__(self, input_groups, merge_dimension):
        self.merge_dimensions = merge_dimension
        self._merges = None
        self._merge_sizes = None
        self._masks = None
        super(MergingInputGroupCombiner, self).__init__(input_groups)

    def update(self):
        dimnames = [x.dimnames for x in self.input_groups.values()]
        sizes = [x.size for x in self.input_groups.values()]

        # Validate the sample dimensions and sizes are consistent
        unique_dimnames = tuple(set(x for dimname in dimnames for x in dimname))
        unique_sizes = {x: set() for x in unique_dimnames}

        for size, dimname in zip(sizes, dimnames):
            for size_element, dimname_element in zip(size, dimname):
                unique_sizes[dimname_element].add(size_element)
        if not all(len(x) == 1 for x in unique_sizes.values()):
            raise exceptions.FastrValueError('The dimension have incosistent sizes: {}'.format(unique_sizes))
        unique_sizes = {k: v.pop() for k, v in unique_sizes.items()}

        # Check how many merges to perform
        if self.merge_dimensions == 'all':
            counts = {name: [sum(x == name for x in dimname) for dimname in dimnames] for name in unique_dimnames}
            merges = {name: min(value) for name, value in counts.items()}
            merges = tuple(key for key, value in merges.items() for x in range(value))
        else:
            merges = tuple(self.merge_dimensions)

        self._merges = merges

        # Make a merge mask for the dimnames
        masks = []
        for dimname in dimnames:
            temp = []
            temp_merges = list(merges)
            for name in dimname:
                if name in temp_merges:
                    index = temp_merges.index(name)
                    temp.append(index)
                    temp_merges[index] = None
                else:
                    temp.append(slice(unique_sizes[name]))
            masks.append(temp)

        self._masks = masks

        # Do the actual merging
        self._merge_sizes = tuple(unique_sizes[x] for x in merges)

    def merge(self, list_of_items):
        new_item = [x for mask, item in zip(self._masks, list_of_items) for m, x in zip(mask, item) if isinstance(m, slice)]
        new_item.extend(x for x, m in zip(list_of_items[0], self._masks[0]) if not isinstance(m, slice))
        return new_item

    def unmerge(self, item):
        index = 0
        result = OrderedDict()
        nr_slices = sum(isinstance(x, slice) for y in self._masks for x in y)
        for input_group, mask in zip(self.input_groups, self._masks):
            original_item = []
            for m in mask:
                if isinstance(m, slice):
                    original_item.append(item[index])
                    index += 1
                else:
                    original_item.append(item[m + nr_slices])
            result[input_group] = tuple(original_item)

        return result

    def iter_input_groups(self):
        # Create iterator for the merged part of the iteration
        if len(self._merges) > 0:
            fixed_indexes = itertools.product(*[range(x) for x in self._merge_sizes])
        else:
            fixed_indexes = [()]

        # Loop over the fixed (merged) indices and then over the remainder
        for fixed_index in fixed_indexes:
            iterators = [itertools.product(*[range(m.stop) if isinstance(m, slice) else [fixed_index[m]] for m in mask]) for mask in self._masks]

            for indexes in itertools.product(*iterators):
                # Retrieve the samples from all input groups
                samples = []

                for index, input_group in zip(indexes, self.input_groups.values()):
                    try:
                        samples.append(input_group[SampleIndex(index)])
                    except exceptions.FastrIndexNonexistent:
                        samples.append(None)

                # Merge all sample payloads
                if all(x is not None for x in samples):
                    yield self.merge_payloads(samples)
                else:
                    fastr.log.info('Skipping {} because it is non-existent due to sparsity!'.format(indexes))




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the object manager class
"""

from abc import abstractproperty, abstractmethod

from ... import exceptions
from ...abc.basemanager import BaseManager
from ...abc.serializable import load
from ...core.version import Version
from ...helpers import log


class ObjectManager(BaseManager):
    """
    Class for managing all the objects loaded in the fastr system
    """
    def __init__(self, path):
        """
        Create a ObjectManager and scan path to search for Objects

        :param path: the path(s) to scan for Objects
        :type path: str or iterable of str
        :return: newly created ObjectManager
        """
        super(ObjectManager, self).__init__(path, True)

    def __contains__(self, key):
        """
        Check if an item is in the ObjectManager

        :param key: object id or tuple (Objectid, version)
        :type key: str or tuple
        :return: flag indicating the item is in the manager
        """
        return self.__keytransform__(key) in self.data

    def __getitem__(self, key):
        """
        Retrieve a Object from the ObjectManager. You can request by only an id,
        which results in the newest version of the Object being returned, or
        request using both an id and a version.

        :param key: object id or tuple (Objectid, version)
        :type key: str or tuple
        :return: the requested Object
        :raises FastrObjectUnknownError: if a non-existing Object was requested
        """
        new_key = self.__keytransform__(key)
        if new_key not in self.data:
            raise exceptions.FastrObjectUnknownError('Key "{}" (expanded to {}) not found in {}'.format(key, new_key, type(self).__name__))

        obj = self.data[new_key]
        return obj

    def __keytransform__(self, key):
        """
        Key transform, used for allowing indexing both by id-only and by
        ``(id, version)``

        :param key: key to transform
        :return: key in form ``(id, version)``
        """
        # Get the version (or None)
        #if not isinstance(key, tuple):
        #    raise exceptions.FastrTypeError('Tool key must be a tuple of (str, tool_version)')

        #if not len(key) == 2:
        #    raise exceptions.FastrValueError('Tool must contain two elements (tool_str, tool_version)')
        #else:
        #    obj_str, obj_version = key

        #if isinstance(obj_str, str):
        #    log.info('Found obj_str: {}'.format(obj_str))
        #    match = re.match(r'^(.*?)/?([^/:]+):([^:]+)$', obj_str)
        #    if not match:
        #        raise exceptions.FastrValueError('Object key should be in form "item:version"')
        #    obj_namespace, object_id, version = match.groups()
        #    version = Version(version)
        #else:
        #    raise exceptions.FastrTypeError('Object key should be a string or tuple!')

        # Check that we are done (should be unique match)
        # return obj_namespace, '{}:{}'.format(object_id, version), version
        obj_id, obj_version = key

        if not isinstance(obj_version, Version):
            obj_version = Version(obj_version)

        return obj_id, obj_version

    @property
    def _item_extension(self):
        """
        Extension of files to load
        """
        return '(xml|json|yaml)'

    @abstractproperty
    def object_class(self):
        """
        The class of the objects to populate the manager with
        """

    @abstractmethod
    def get_object_version(self, obj):
        """
        Get the version of a given object

        :param object: the object to use
        :return: the version of the object
        """

    def _print_key(self, key):
        """
        Print function for the keys
        """
        return key[0], str(key[1])

    def _print_value(self, value):
        """
        Print function for the values
        """
        return value.filename

    def _load_item(self, filepath, namespace):
        """
        Load a Object file and store it in the Manager
        """

        obj = load(filepath, cls=self.object_class)
        object_version = self.get_object_version(obj)

        # Make sure the last part of the directory structure is not the version
        # if that is the case, strip it
        if len(namespace) > 0:
            possible_version = namespace[-1]

            try:
                possible_version = Version(possible_version)
                if possible_version == object_version:
                    namespace = namespace[:-1]
            except exceptions.FastrVersionInvalidError:
                pass

        namespace = '/'.join(namespace)
        obj.namespace = namespace

        self._store_item((obj.ns_id, obj.version), obj)

    def todict(self):
        """
        Return a dictionary version of the Manager

        :return: manager as a dict
        """
        result = {}
        for key in self.keys():
            if key[0] not in result:
                result[key[0]] = []

            if str(key[1]) not in result[key[0]]:
                result[key[0]].append(str(key[1]))

        return result

    def objectversions(self, obj):
        """
        Return a list of available versions for the object

        :param object: The object to check the versions for. Can be either a `Object` or a `str`.
        :return: List of version objects. Returns `None` when the given object is not known.
        """

        if isinstance(obj, self.object_class):
            ns_id = obj.ns_id
        elif isinstance(obj, str):
            ns_id = obj

        log.info("Object namespaced id: {}".format(ns_id))

        return sorted(key[1] for key in self.keys() if key[0] == ns_id)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the tool manager class
"""

from .objectmanager import ObjectManager
from .iopluginmanager import IOPluginManager
from ...core.tool import Tool
from ... import resources


class ToolManager(ObjectManager):
    def populate(self):
        super().populate()
        IOPluginManager.create_ioplugin_tool(self, interfaces=resources.interfaces)

    @property
    def object_class(self):
        return Tool

    def get_object_version(self, obj):
        return obj.command['version']

    def toolversions(self, tool):
        """
        Return a list of available versions for the tool

        :param tool: The tool to check the versions for. Can be either a `Tool` or a `str`.
        :return: List of version objects. Returns `None` when the given tool is not known.
        """
        return self.objectversions(tool)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The module containing the classes describing the targets.
"""

import os
import subprocess

import fastr
from fastr import exceptions
from fastr.core.target import SubprocessBasedTarget


class SingularityTarget(SubprocessBasedTarget):
    """
    A tool target that is run using a singularity container,
    see the `singulary website <http://singularity.lbl.gov/>`_

    * ``binary (required)``: the name of the binary/script to call, can also be called ``bin``
      for backwards compatibility.
    * ``container (required)``: the singularity container to run, this can be in url form for singularity
                                pull or as a path to a local container

    * ``interpreter``: the interpreter to use to call the binary e.g. ``python``

    """
    SINGULARITY_BIN = 'singularity'

    def __init__(self, binary, container, interpreter=None):
        """
        Define a new local binary target. Must be defined either using paths and optionally environment_variables
        and initscripts, or enviroment modules.
        """
        self.binary = binary
        self.container = container
        self.interpreter = interpreter

    def __enter__(self):
        """
        Set the environment in such a way that the target will be on the path.
        """
        super(SingularityTarget, self).__enter__()

        # TODO Make sure the container is present, otherwise download the container
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Cleanup the environment
        """
        # TODO We could potentially remove the container again if we downloaded it in the first place?
        pass

    @classmethod
    def test(cls):
        """
        Test if singularity is availble on the path
        """
        try:
            subprocess.check_output([cls.SINGULARITY_BIN, '--help'], stderr=subprocess.STDOUT)
        except OSError:
            raise exceptions.FastrExecutableNotFoundError(cls.SINGULARITY_BIN)

    def run_command(self, command):
        # Add interpreter if needed
        if self.interpreter is not None:
            command = [self.interpreter] + command

        # Bind all fastr mounts
        mounts = fastr.config.mounts.values()
        binds = []
        for mount in mounts:
            if not os.path.exists(mount):
                continue

            binds.append('--bind')
            binds.append('{x}:{x}'.format(x=mount))

        fastr.log.info('Singularity binds: {}'.format(binds))

        # Compose the singularity command
        singularity_command = [self.SINGULARITY_BIN, 'exec']
        singularity_command.extend(binds)
        singularity_command.append(self.container)
        singularity_command.extend(command)

        # Create final command
        fastr.log.debug('Command: {}'.format(command))
        fastr.log.debug('Singularity command: {}'.format(singularity_command))
        return self.call_subprocess(singularity_command)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module to maintain a tool.

Exported classes:

* Tool -- A class encapsulating a tool.
* ParameterDescription -- The base class containing the shared description of
  a parameter (both input and ouput).
* InputParameterDescription -- A class containing the description of an input
  parameter.
* Output ParameterDescription -- A class containing the description of an
  output parameter.
"""

import itertools
import os
import platform
import re
import shutil
from collections import namedtuple, OrderedDict
from tempfile import mkdtemp

import jsonschema

from . import vfs_plugin
from .cardinality import create_cardinality
from .interface import InterfaceResult
from .version import Version
from .. import exceptions
from ..abc.baseplugin import PluginState
from ..abc.serializable import Serializable
from ..datatypes import types
from ..helpers import log, config
from ..helpers.checksum import hashsum
from ..helpers import iohelpers
from .. import resources


class Tool(Serializable):
    """
    The class encapsulating a tool.
    """

    __dataschemafile__ = 'Tool.schema.json'

    test_spec = namedtuple('TestSpecification', ['input', 'command', 'output'])
    TOOL_REFERENCE_FILE_NAME = '__fastr_tool_ref__.json'
    TOOL_RESULT_FILE_NAME = '__fastr_tool_result.pickle.gz'

    DEFAULT_TARGET_CLASS = {
        'MacroNode': 'MacroTarget'
    }

    def __init__(self, doc=None):
        """
        Create a new Tool
        :param doc: path of toolfile or a dict containing the tool data
        :type doc: str or dict
        """

        # Cache value for target binary
        self._target = None

        if doc is None:
            return

        filename = None
        if not isinstance(doc, dict):
            log.debug('Trying to load file: {}'.format(doc))
            filename = os.path.expanduser(doc)
            filename = os.path.abspath(filename)
            doc = self._loadf(filename)

        # Check if the doc is a valid Tool structure
        try:
            # __unserializer__ is supplied by the serializable meta-class
            # pylint: disable=no-member
            doc = Tool.get_serializer().instantiate(doc)
        except jsonschema.ValidationError:
            log.error('Could not validate Tool data again the schema!')
            raise
        else:
            log.debug('Tool schema validated!')

        # Get attributes from root node
        self.filename = filename

        #: Identifier for the tool
        regex = r'^[A-Z][\w\d_]*$'
        if re.match(regex, doc['id']) is None:
            message = 'A tool id in Fastr should be UpperCamelCase as enforced' \
                      ' by regular expression {} (found {})'.format(regex, doc['id'])
            log.warning(message)

        self._id = doc['id']

        #: The namespace this tools lives in, this will be set by the ToolManager on load
        self.namespace = None

        #: Name of the tool, this should be a descriptive, human readable name.
        self.name = doc.get('name', self._id)

        #: Version of the tool, not of the underlying software
        self.version = Version(str(doc.get('version')))

        #: Class for of the Node to use
        self.node_class = doc.get('class', 'Node')

        if self._id is None or self.name is None or self.version is None:
            raise exceptions.FastrValueError('Tool should contain an id, name and version!')

        #: URL to website where this tool can be downloaded from
        self.url = doc.get('url', '')

        #: Description of the tool and it's functionality
        self.description = doc.get('description', '')

        #: List of authors of the tool. These people wrapped the executable but
        #: are not responsible for executable itself.
        self.authors = doc['authors']

        #: List of tags for this tool
        self.tags = doc.get('tags', [])

        # Parse references field and format them into a dictionary
        #: A list of documents and in depth reading about the methods used in this tool
        self.references = doc.get('references', [])

        #: Requirements for this Tool
        #:
        #: .. warning:: Not yet implemented
        self.requirements = None

        #: Test for this tool. A test should be a collection of inputs, parameters
        #: and outputs to verify the proper functioning of the Tool.
        #:
        #: The format of the tests is a list of namedtuples, that have 3 fields:
        #: - input: a dict of the input data
        #: - command: a list given the expected command-line arguments
        #: - output: a dict of the output data to validate
        #:
        #: .. warning:: Not yet implemented

        self.tests = doc['tests']

        command = doc['command']

        # Find commands
        #: Command is a dictionary contain information about the command which is
        #: called by this Tool:
        #: command['interpreter'] holds the (possible) interpreter to use
        #: command['targets'] holds a per os/arch dictionary of files that should be executed
        #: command['url'] is the webpage of the command to be called
        #: command['version'] is the version of the command used
        #: command['description'] can help a description of the command
        #: command['authors'] lists the original authors of the command
        self.command = {
            'targets': command['targets'],
            'license': command.get('license', ''),
            'url': command.get('url', ''),
            'version': Version(str(command.get('version'))),
            'description': command.get('description', ''),
            'authors': command.get('authors', [])
        }

        if len(self.command['targets']) == 0:
            raise exceptions.FastrValueError("No targets defined in tool description.")

        #: This holds the citation you should use when publishing something based on this Tool
        self.cite = doc['cite']

        #: Man page for the Tool. Here usage and examples can be described in detail
        self.help = doc['help']

        #: Create the Interface based on the class specified in the tool file
        interface_class = resources.interfaces[doc['interface'].get('class', 'FastrInterface')]

        if interface_class.status != PluginState.loaded:
            raise exceptions.FastrPluginNotLoaded(
                'Required InterfacePlugin {} was not loaded properly (status {})'.format(interface_class.id,
                                                                                         interface_class.status)
            )

        self.interface = interface_class(id_='{}_{}_interface'.format(self.id, self.command['version']),
                                         document=doc['interface'])

    @property
    def hash(self):
        return hashsum(self.__getstate__())

    @property
    def id(self):
        return '{}:{}'.format(self._id, self.command['version'])

    @property
    def ns_id(self):
        """
        The namespace and id of the Tool
        """
        if self.namespace is None:
            return self.id
        else:
            return '{}/{}'.format(self.namespace, self.id)

    @property
    def fullid(self):
        """
        The full id of this tool
        """
        return 'fastr:///tools/{}/{}'.format(self.ns_id, self.version)

    @property
    def inputs(self):
        return self.interface.inputs

    @property
    def outputs(self):
        return self.interface.outputs

    @property
    def target(self):
        """
        The OS and arch matched target definition.
        """
        if self._target is not None:
            return self._target

        # Get platform and os
        arch = platform.architecture()[0].lower()
        os_ = platform.system().lower()

        matching_targets = [x for x in self.command['targets'] if x['os'] in [os_, '*'] and x['arch'] in [arch, '*']]

        if len(matching_targets) == 0:
            return None
        elif len(matching_targets) == 1:
            target = matching_targets[0]
        else:
            # This should give the optimal match
            for match in matching_targets:
                match['score'] = 0

                if match['os'] == os_:
                    match['score'] += 2

                if match['arch'] == arch:
                    match['score'] += 1

            matching_targets = sorted(matching_targets, reverse=True, key=lambda x: x['score'])
            log.debug('Sorted matches: {}'.format(matching_targets))
            target = matching_targets[0]

        # Make sure target is a copy
        target = dict(target)
        cls = self.DEFAULT_TARGET_CLASS.get(self.node_class)
        if cls is None:
            cls = target.pop('class', 'LocalBinaryTarget')
        cls = resources.targets[cls]

        # Create target from curdir
        old_curdir = os.path.abspath(os.curdir)
        os.chdir(self.path)

        # Check if the argument binary exists (or use bin as alternative)
        if 'binary' not in target and 'bin' in target:
            target['binary'] = target.pop('bin')

        # Remove needless fields
        target.pop('os', None)
        target.pop('arch', None)

        # Instantiate target
        try:
            self._target = cls(**target)
        except Exception as exception:
            log.error('Could not load tool {} because the following exception was encountered: {}'.format(
                self.ns_id, exception
            ))
            raise

        os.chdir(old_curdir)

        return self._target

    def _prepare_payload(self, payload=None, **kwargs):
        # Allow kwargs to be used instead of payload
        if payload is None:
            payload = {'inputs': {}, 'outputs': {}}
            for key, value in kwargs.items():
                if key in self.inputs and key in self.outputs:
                    raise exceptions.FastrValueError(
                        'Cannot figure out if "{}" is an input or output, please prefix with input_/output_ as needed'
                    )
                elif key in self.inputs:
                    payload['inputs'][key] = value
                elif key in self.outputs:
                    payload['outputs'][key] = value
                elif key.startswith('input_') and key[6:] in self.inputs:
                    payload['inputs'][key[6:]] = value
                elif key.startswith('output_') and key[7:] in self.outputs:
                    payload['outputs'][key[7:]] = value
                else:
                    raise exceptions.FastrValueError('Cannot match key "{}" to any input/output!'.format(key))

        # Make sure all values are wrapped in a tuple (for single values)
        for key, value in payload['inputs'].items():
            if not isinstance(value, (tuple, OrderedDict)):
                payload['inputs'][key] = (value,)
        for key, value in payload['outputs'].items():
            if not isinstance(value, (tuple, OrderedDict)):
                payload['outputs'][key] = (value,)

        return payload

    def _validate_payload(self, payload):
        for input_ in self.inputs.values():
            if input_.id not in payload['inputs']:
                if input_.required:
                    raise exceptions.FastrValueError('Could not find data for required input {}'.format(input_.id))
                else:
                    continue

            value = payload['inputs'][input_.id]

            if not isinstance(value, (tuple, OrderedDict)):
                raise exceptions.FastrTypeError('Value needs to be a tuple or OrderedDict')

            cardinality_spec = create_cardinality(input_.cardinality, None)
            if not (not input_.required and len(value) == 0) and \
                    not cardinality_spec.validate(payload, len(value)):
                raise exceptions.FastrValueError(
                    'Input cardinality mismatch for {}. Cardinality "{}" does not match "{!r}"'.format(
                        input_.id,
                        len(value),
                        cardinality_spec,
                    )
                )

            datatype = types.get(input_.datatype)

            # If the input has named sub-inputs we need to collapse the structure
            # before checking the datatypes
            if isinstance(value, OrderedDict):
                value = (x for y in value.values() for x in y)

            if datatype is not None and not all(datatype.isinstance(x) for x in value):
                raise exceptions.FastrTypeError('Input {} has an incorrect datatype (found {}, expected {})'.format(
                    input_.id,
                    set(type(x) for x in value),
                    datatype
                ))

    def execute(self, payload=None, **kwargs) -> InterfaceResult:
        """
        Execute a Tool given the payload for a single run

        :param payload: the data to execute the Tool with
        :returns: The result of the execution
        :rtype: InterFaceResult
        """
        payload = self._prepare_payload(payload=payload, **kwargs)

        # Make sure hte payload is valid
        self._validate_payload(payload)

        # Set environment variables if not set
        os.environ.setdefault('FASTR_SAMPLE_ID', 'UNKNOWN')
        os.environ.setdefault('FASTR_SAMPLE_INDEX', '0')
        os.environ.setdefault('FASTR_JOB_ID', 'fastr_interactive_0')
        os.environ.setdefault('FASTR_WORK_DIR', 'UNKNOWN')

        target = self.target
        log.info('Target is {}'.format(target))

        if target is None:
            arch = platform.architecture()[0].lower()
            os_ = platform.system().lower()
            raise exceptions.FastrValueError(
                'Cannot find a viable target for {}/{} on {} ({} bit)'.format(self.id, self.version, os_, arch)
            )

        log.info('Using payload: {}'.format(payload))
        with target:
            result = self.interface.execute(target, payload)

        return result

    def __str__(self):
        """
        Get a string version for the Tool

        :return: the string version
        :rtype: str
        """
        return '<Tool: {} version: {}>'.format(self.id, self.version)

    def __repr__(self):
        """
        Get a string representation for the Tool. This will show the inputs
        and output defined in a table-like structure.

        :return: the string representation
        :rtype: str
        """
        if self.name is not None and len(self.name) > 0:
            name_part = ' ({})'.format(self.name)
        else:
            name_part = ''

        return_list = ["Tool {} v{}{}".format(self.id, str(self.command['version']), name_part)]

        # The "+ [8]" guarantees a minimum of 8 width and avoids empty lists
        width_input_keys = max([len(x.id) for x in self.inputs.values()] + [8])
        width_input_types = max([len(x.datatype) for x in self.inputs.values()] + [8]) + 2
        width_output_keys = max([len(x.id) for x in self.outputs.values()] + [8])
        width_output_types = max([len(x.datatype) for x in self.outputs.values()] + [8]) + 2

        return_list.append('{:^{}}  | {:^{}}'.format('Inputs', width_input_types + width_input_keys + 1,
                                                     'Outputs', width_output_types + width_output_keys + 1))
        return_list.append('-' * (width_input_keys + width_input_types + width_output_keys + width_output_types + 7))
        for input_, output in itertools.zip_longest(self.inputs.values(), self.outputs.values()):
            if input_ is None:
                input_id = ''
                input_type = ''
            else:
                input_id = input_.id
                input_type = '({})'.format(input_.datatype)

            if output is None:
                output_id = ''
                output_type = ''
            else:
                output_id = output.id
                output_type = '({})'.format(output.datatype)

            return_list.append('{:{}} {:{}}  |  {:{}} {:{}}'.format(input_id, width_input_keys,
                                                                    input_type, width_input_types,
                                                                    output_id, width_output_keys,
                                                                    output_type, width_output_types))

        return '\n'.join(return_list)

    def __eq__(self, other):
        """Compare two Tool instances with each other.

        :param other: the other instances to compare to
        :type other: Tool
        :returns: True if equal, False otherwise
        """
        if not isinstance(other, Tool):
            return NotImplemented

        dict_self = dict(self.__dict__)
        del dict_self['_target']
        del dict_self['namespace']
        del dict_self['filename']

        dict_other = dict(other.__dict__)
        del dict_other['_target']
        del dict_other['namespace']
        del dict_other['filename']

        return dict_self == dict_other

    def __getstate__(self):
        """
        Retrieve the state of the Tool

        :return: the state of the object
        :rtype dict:
        """
        state = dict(vars(self))
        state['command'] = {k: v for k, v in list(self.command.items())}

        state['class'] = state['node_class']
        state['interface'] = self.interface.__getstate__()
        state['command']['version'] = str(self.command['version'])
        state['version'] = str(self.version)
        state['id'] = self._id
        del state['node_class']
        del state['_target']

        return state

    def serialize(self):
        """
        Prepare data for serialization, this removes some fields from the state
        that are not needed when serializing to a file
        """
        state = self.__getstate__()

        # Data stub to be ordered
        data = {
            "id": self._id,
            "name": self.name,
            "version": str(self.version),
            "description": self.description,
            "authors": self.authors,
            "class": state['class'],
            "command": state['command'],
            "interface": state['interface'],
        }

        # Set of items to skip
        skip = {'_id', 'namespace', 'filename'}

        for key, value in state.items():
            if key in skip or key in data:
                continue

            data[key] = value

        return data

    def __setstate__(self, state):
        """
        Set the state of the Tool by the given state.

        :param dict state: The state to populate the object with
        """
        if 'filename' not in state:
            state['filename'] = None

        if 'namespace' not in state:
            state['namespace'] = None

        if 'id' in state:
            state['_id'] = state.pop('id')
            
        state['version'] = Version(state['version'])
        state['command']['version'] = Version(str(state['command']['version']))

        state['node_class'] = state['class']
        del state['class']

        interface_class = resources.interfaces[state['interface'].get('class', 'FastrInterface')]
        if 'id' not in state['interface']:
            state['interface']['id'] = '{}_{}_interface'.format(state['_id'], state['command']['version'])

        if interface_class.status != PluginState.loaded:
            raise exceptions.FastrPluginNotLoaded(
                'Required InterfacePlugin {} was not loaded properly (status {})'.format(interface_class.id,
                                                                                         interface_class.status)
            )

        state['interface'] = interface_class.deserialize(state['interface'])
        self.__dict__.update(state)
        self._target = None

    @property
    def path(self):
        """
        The path of the directory in which the tool definition file was
        located.
        """
        return os.path.dirname(self.filename)

    @property
    def command_version(self):
        return self.command['version']

    def create_reference(self, input_data, output_directory, mount_name='__ref_tmp__', 
                         copy_input=True):
        # Make sure the output directory is absolute for later on
        output_directory = os.path.abspath(os.path.expanduser(output_directory))

        # Create output directory
        if not os.path.exists(output_directory):
            os.makedirs(output_directory)

        input_data_dir = os.path.join(output_directory, 'input_data')

        if not os.path.exists(input_data_dir):
            os.mkdir(input_data_dir)

        output_data_dir = os.path.join(output_directory, 'output_data')

        if not os.path.exists(output_data_dir):
            os.mkdir(output_data_dir)

        # Add the temporary mount
        config.read_config_string('mounts["{}"] = "{}"'.format(mount_name, output_directory))

        # If desired, copy input data to data directory in output directory
        if copy_input:
            log.info('Copying input data into reference directory...')
            new_input_data = {}
            stored_input_data = {}
            for input_id, input_value in input_data.items():
                new_input_value = []
                stored_input_value = []

                # Make sure we are working with tuples
                if isinstance(input_value, list):
                    input_value = tuple(input_value)

                if not isinstance(input_value, tuple):
                    input_value = input_value,

                # Datatype for this input
                datatype = types[self.inputs[input_id].datatype]

                # If needed copy the data
                for value in input_value:
                    if isinstance(value, str) and os.path.exists(value):
                        filename = os.path.basename(value)
                        destination = 'vfs://{}/input_data/{}'.format(mount_name, filename)
                        vfs_plugin.push_sink_data(value, destination, datatype=datatype)
                        new_input_value.append(vfs_plugin.url_to_path(destination))
                        stored_input_value.append('$REFDIR/input_data/{}'.format(filename))
                    else:
                        new_input_value.append(value)
                        stored_input_value.append(value)
                new_input_data[input_id] = tuple(new_input_value)
                stored_input_data[input_id] = tuple(stored_input_value)
        else:
            new_input_data = input_data
            stored_input_data = input_data

        test_data = {
            'input_data': stored_input_data,
            'used_input_data': new_input_data,
            'original_input_data': input_data,
            'tool': self.ns_id,
            'version': str(self.command_version),
        }

        # New replace input_data with new input data
        input_data = new_input_data

        # Save the test summary to the reference file
        log.info('Saving basic reference information')
        iohelpers.save_json(
            os.path.join(output_directory, self.TOOL_REFERENCE_FILE_NAME),
            test_data
        )

        # Cast input data to the desired datatype
        for input_id, input_value in input_data.items():
            _datatype = self.inputs[input_id].datatype
            input_data[input_id] = tuple(
                    [types[_datatype](input_value[idx]) 
                     for idx, val in enumerate(input_value)]
                )

        # Reference payload as a shortcut
        payload = {'inputs': input_data, 'outputs': {}}

        # Create output arguments automatically
        from fastr.execution.job import Job

        log.info('Determining values for outputs')
        for id_, argument in self.outputs.items():
            # Determine cardinality
            if not argument.automatic:
                cardinality = create_cardinality(argument.cardinality, None)
                cardinality = cardinality.calculate_job_cardinality(payload)
            else:
                cardinality = 1

            # Create output arguments automatically
            payload['outputs'][id_] = Job.fill_output_argument(output_spec=argument,
                                                               cardinality=cardinality,
                                                               desired_type=types[argument.datatype],
                                                               requested=True,
                                                               tmpurl='vfs://{}/output_data'.format(mount_name))

        log.info('Resulting test data: {}'.format(test_data))

        log.info('Executing tool...')
        result = self.execute(payload=payload)
        result.command = result.target_result.command

        # Translate command to avoid hard links to output_directory
        log.info('Processing results for use as reference data')

        # Do no substitute for input data
        pattern = r'^{}'.format(output_data_dir)
        result.reference_command = [re.sub(pattern, '$OUTDIR', x, 1) for x in result.target_result.command]
        result.reference_command = ['$INDIR/{}'.format(os.path.basename(x)) if os.path.exists(x) else x
                                    for x in result.reference_command]

        # Translate results back
        output_data = {}
        for key, value in result.result_data.items():
            datatype = types[self.outputs[key].datatype]
            preferred_type = (datatype,)

            output_data[key] = Job.translate_output_results(value,
                                                            datatypes=preferred_type,
                                                            mountpoint=mount_name)
        
        result.output_data = output_data

        config.read_config_string('del mounts["{}"]'.format(mount_name))

        log.info('Saving results for reference...')
        iohelpers.save_gpickle(
            os.path.join(output_directory, self.TOOL_RESULT_FILE_NAME),
            result
        )
        log.info('Finishing creating reference data for {}/{}'.format(self.ns_id, self.command_version))

    @staticmethod
    def compare_output_data(current_output_data, reference_output_data, validation_result, output):
        log.info('Current output data: {}'.format(current_output_data))
        log.info('Reference output data: {}'.format(reference_output_data))
        for nr, (current_value, reference_value) in enumerate(zip(current_output_data, reference_output_data)):
            if current_value != reference_value:
                validation_result.append('\n'.join((
                    'Value for {}/{} was not equal! (found "{}", expected "{}")'.format(
                        output,
                        nr,
                        current_value,
                        reference_value,
                    ),
                    'Output: [{}] {!r}'.format(type(current_value).__name__,
                                               current_value),
                    'Reference: [{}] {!r}'.format(type(reference_value).__name__,
                                                  reference_value),
                )))

    @classmethod
    def test_tool(cls, reference_data_dir, tool=None, input_data=None):
        """
        Execute the tool with the input data specified and test the results
        against the refence data. This effectively tests the tool execution.

        :param str reference_data_dir: The path or vfs url of reference data to compare with
        :param dict source_data: The source data to use
        """
        if not isinstance(reference_data_dir, str):
            raise exceptions.FastrTypeError('reference_data_dir should be a string!')

        if reference_data_dir.startswith('vfs://'):
            reference_data_dir = vfs_plugin.url_to_path(reference_data_dir)

        if not os.path.isdir(reference_data_dir):
            raise exceptions.FastrTypeError('The reference_data_dir should be pointing to an existing directory!'
                                            ' {} does not exist'.format(reference_data_dir))

        test_data = iohelpers.load_json(
            os.path.join(reference_data_dir, cls.TOOL_REFERENCE_FILE_NAME)
        )

        if tool is None:
            tool = resources.tools[test_data['tool'], test_data['version']]

        log.info('Testing tool {}/{} against {}'.format(
            tool.ns_id,
            tool.command_version,
            reference_data_dir,
        ))

        if input_data is None:
            input_data = {}

            for key, value in test_data['input_data'].items():
                if not isinstance(value, (tuple, list)):
                    value = value,

                # Set the $REFDIR correctly (the avoid problems with moving the reference dir)
                value = tuple(x.replace('$REFDIR', reference_data_dir) if isinstance(x, str) else x for x in value)
                input_data[key] = value

        temp_results_dir = None
        reference_result = iohelpers.load_gpickle(
            os.path.join(reference_data_dir, cls.TOOL_RESULT_FILE_NAME)
            )

        validation_result = []
        try:
            # Create temporary output directory
            temp_results_dir = os.path.normpath(mkdtemp(
                prefix='fastr_tool_test_{}_'.format(tool.id), dir=config.mounts['tmp']
            ))

            # Create a new reference for comparison
            log.info('Running tool and creating new reference data for comparison...')
            try:
                tool.create_reference(input_data,
                                      temp_results_dir,
                                      mount_name='__test_tmp__',
                                      copy_input=False)
            except Exception as exception:
                log.warning('Encountered exception when trying to run the {}/{} tool!'.format(
                    tool.ns_id, tool.command_version)
                )
                log.warning('Exception: [{}] {}'.format(type(exception).__name__, exception))
                validation_result.append('Encountered {}: {}'.format(type(exception).__name__, exception))
                return validation_result

            current_result = iohelpers.load_gpickle(os.path.join(temp_results_dir, cls.TOOL_RESULT_FILE_NAME))

            log.info('Comparing current run against reference run...')

            # First check the command and return code
            if current_result.reference_command != reference_result.reference_command:
                validation_result.append('\n'.join((
                    'Different command used for execution of tool "{}/{}"'.format(tool.ns_id, tool.command_version),
                    'Current command: {}'.format(current_result.reference_command),
                    'Reference command {}'.format(reference_result.reference_command),
                )))

            if current_result.target_result.return_code != reference_result.target_result.return_code:
                validation_result.append('\n'.join((
                    'Different returncode used for execution of tool "{}/{}"'.format(tool.ns_id,
                                                                                     tool.returncode_version),
                    'Current returncode: {}'.format(current_result.target_result.returncode),
                    'Reference returncode {}'.format(reference_result.target_result.returncode),
                )))

            # Check if the outputs are the same
            current_outputs = sorted(reference_result.output_data.keys())
            reference_outputs = sorted(current_result.output_data.keys())
            if current_outputs != reference_outputs:
                validation_result.append('\n'.join((
                    'Different outputs found in tool "{}/{}"'.format(tool.ns_id, tool.command_version),
                    'Current outputs: {}'.format(current_outputs),
                    'Reference outputs {}'.format(reference_outputs),
                )))

            # Add the mounts need to find the data
            config.read_config_string('mounts["__test_tmp__"] = "{}"'.format(temp_results_dir))
            config.read_config_string('mounts["__ref_tmp__"] = "{}"'.format(reference_data_dir))

            # Check if values for all outputs match
            for output in reference_outputs:
                current_output_data = current_result.output_data[output]
                reference_output_data = reference_result.output_data[output]
                for idx, _ in enumerate(current_output_data):
                    if not isinstance(current_output_data[idx], 
                                      type(reference_output_data[idx])):
                        validation_result.append('\n'.join((
                            'Different type for output {} found in tool "{}/{}"'.format(
                                tool.ns_id, tool.command_version
                            ),
                            'Current type: {}'.format(type(current_output_data[idx])),
                            'Reference type {}'.format(type(reference_output_data[idx])),
                        )))
                if isinstance(current_output_data, dict):
                    raise exceptions.FastrNotImplementedError(
                        'Output comparison of dict structures is not supported yet!'
                    )

                else:
                    cls.compare_output_data(
                        current_output_data,
                        reference_output_data,
                        validation_result,
                        output
                    )

            # Remove the temporary mounts again
            config.read_config_string('del mounts["__test_tmp__"]')
            config.read_config_string('del mounts["__ref_tmp__"]')

            if len(validation_result) == 0:
                log.info('Run and reference were equal! Test passed!')
            else:
                log.info('Found difference with reference data! Test failed!')
                for line in validation_result:
                    log.info(line)
            return validation_result
        finally:
            # Clean up
            log.info('Removing temp result directory {}'.format(temp_results_dir))
            if temp_results_dir is not None and os.path.isdir(temp_results_dir):
                shutil.rmtree(temp_results_dir, ignore_errors=True)

    def test(self, reference=None):
        """
        Run the tests for this tool
        """
        if reference is None:
            result = []
            tool_dir = os.path.dirname(self.filename)
            for test in self.tests:
                reference_dir = os.path.abspath(os.path.join(tool_dir, test))
                try:
                    result.extend(self.test_tool(reference_data_dir=reference_dir, 
                                                 tool=self))
                except exceptions.FastrTypeError:
                    message = 'Reference data in {} is not valid!'.format(reference_dir)
                    log.warning(message)
                    result.append(message)
        else:
            result = self.test_tool(reference_data_dir=reference, tool=self)

        return result



import argparse
import os
import subprocess

from fastr.utils.cmd import add_parser_doc_link


def get_parser():
    parser = argparse.ArgumentParser(
        description="Run pylint in such a way that the output is written to a file")
    parser.add_argument('--output_file', metavar='PYLINT.OUT', required=True, type=str,
                        help='The file to result in')
    return parser


def run_pylint(out_file, pylint_args):
    print('Saving result to {}'.format(out_file))
    with open(out_file, 'w') as fh_out:
        print('Running: pylint {}'.format(' '.join(pylint_args)))
        proc = subprocess.Popen(['pylint'] + pylint_args, stdout=fh_out)
        proc.wait(timeout=600)  # Max 10 minutes


def main():
    """
    Tiny wrapper in pylint so the output can be saved to a file (for test automation)
    """
    parser = add_parser_doc_link(get_parser(), __file__)
    args, unknown = parser.parse_known_args()
    output_file = os.path.abspath(args.output_file)
    run_pylint(output_file, unknown)


if __name__ == '__main__':
    main()
